# Copyright (c) 2019-2021, Manfred Moitzi
# License: MIT License
from functools import partial
from ezdxf.lldxf.hdrvars import SingleValue, Point2D, Point3D, HeaderVarDef
from ezdxf.lldxf.const import (
    DXF12, DXF2000, DXF2004, DXF2007, DXF2010, DXF2013, DXF2018,
)

HEADER_VAR_MAP = {
    '$ACADVER': HeaderVarDef(
        name='$ACADVER',
        code=1,
        factory=partial(SingleValue, code=1),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=0,
        default='AC1032'),
    '$ACADMAINTVER': HeaderVarDef(
        name='$ACADMAINTVER',
        # group code changed to 90 in DXF R2018+, this fact is handled in:
        # ezdxf.sections.header.HeaderSection.export_dxf()
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=100,
        default=4),
    '$DWGCODEPAGE': HeaderVarDef(
        name='$DWGCODEPAGE',
        code=3,
        factory=partial(SingleValue, code=3),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=200,
        default='ANSI_1252'),
    '$LASTSAVEDBY': HeaderVarDef(
        name='$LASTSAVEDBY',
        code=1,
        factory=partial(SingleValue, code=1),
        mindxf=DXF2004,
        maxdxf=DXF2018,
        priority=300,
        default='ezdxf'),
    '$REQUIREDVERSIONS': HeaderVarDef(
        name='$REQUIREDVERSIONS',
        code=160,
        factory=partial(SingleValue, code=160),
        mindxf=DXF2013,
        maxdxf=DXF2018,
        priority=400,
        default=0),
    '$INSBASE': HeaderVarDef(
        name='$INSBASE',
        code=10,
        factory=Point3D,
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=500,
        default=(0.0, 0.0, 0.0)),
    '$EXTMIN': HeaderVarDef(
        name='$EXTMIN',
        code=10,
        factory=Point3D,
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=600,
        default=(-100, -100, -100)),
    '$EXTMAX': HeaderVarDef(
        name='$EXTMAX',
        code=10,
        factory=Point3D,
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=700,
        default=(100, 100, 100)),
    '$LIMMIN': HeaderVarDef(
        name='$LIMMIN',
        code=10,
        factory=Point2D,
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=800,
        default=(0.0, 0.0)),
    '$LIMMAX': HeaderVarDef(
        name='$LIMMAX',
        code=10,
        factory=Point2D,
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=900,
        default=(420.0, 297.0)),
    '$ORTHOMODE': HeaderVarDef(
        name='$ORTHOMODE',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=1000,
        default=0),
    '$REGENMODE': HeaderVarDef(
        name='$REGENMODE',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=1100,
        default=1),
    '$FILLMODE': HeaderVarDef(
        name='$FILLMODE',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=1200,
        default=1),
    '$DRAGMODE': HeaderVarDef(
        name='$DRAGMODE',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF12,
        priority=1250,
        default=2),
    '$QTEXTMODE': HeaderVarDef(
        name='$QTEXTMODE',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=1300,
        default=0),
    '$MIRRTEXT': HeaderVarDef(
        name='$MIRRTEXT',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=1400,
        default=1),
    '$OSMODE': HeaderVarDef(
        name='$OSMODE',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF12,
        priority=1400,
        default=20583),
    '$LTSCALE': HeaderVarDef(
        name='$LTSCALE',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=1500,
        default=1.0),
    '$ATTMODE': HeaderVarDef(
        name='$ATTMODE',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=1600,
        default=1),
    '$TEXTSIZE': HeaderVarDef(
        name='$TEXTSIZE',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=1700,
        default=2.5),
    '$TRACEWID': HeaderVarDef(
        name='$TRACEWID',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=1800,
        default=1.0),
    '$TEXTSTYLE': HeaderVarDef(
        name='$TEXTSTYLE',
        code=7,
        factory=partial(SingleValue, code=7),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=1900,
        default='Standard'),
    '$CLAYER': HeaderVarDef(
        name='$CLAYER',
        code=8,
        factory=partial(SingleValue, code=8),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=2000,
        default='0'),
    '$CELTYPE': HeaderVarDef(
        name='$CELTYPE',
        code=6,
        factory=partial(SingleValue, code=6),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=2100,
        default='ByLayer'),
    '$CECOLOR': HeaderVarDef(
        name='$CECOLOR',
        code=62,
        factory=partial(SingleValue, code=62),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=2200,
        default=256),
    '$CELTSCALE': HeaderVarDef(
        name='$CELTSCALE',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=2300,
        default=1.0),
    '$DISPSILH': HeaderVarDef(
        name='$DISPSILH',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=2400,
        default=0),
    '$DIMSCALE': HeaderVarDef(
        name='$DIMSCALE',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=2500,
        default=1.0),
    '$DIMASZ': HeaderVarDef(
        name='$DIMASZ',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=2600,
        default=2.5),
    '$DIMEXO': HeaderVarDef(
        name='$DIMEXO',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=2700,
        default=0.625),
    '$DIMDLI': HeaderVarDef(
        name='$DIMDLI',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=2800,
        default=3.75),
    '$DIMRND': HeaderVarDef(
        name='$DIMRND',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=2900,
        default=0.0),
    '$DIMDLE': HeaderVarDef(
        name='$DIMDLE',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=3000,
        default=0.0),
    '$DIMEXE': HeaderVarDef(
        name='$DIMEXE',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=3100,
        default=1.25),
    '$DIMTP': HeaderVarDef(
        name='$DIMTP',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=3200,
        default=0.0),
    '$DIMTM': HeaderVarDef(
        name='$DIMTM',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=3300,
        default=0.0),
    '$DIMTXT': HeaderVarDef(
        name='$DIMTXT',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=3400,
        default=2.5),
    '$DIMCEN': HeaderVarDef(
        name='$DIMCEN',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=3500,
        default=2.5),
    '$DIMTSZ': HeaderVarDef(
        name='$DIMTSZ',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=3600,
        default=0.0),
    '$DIMTOL': HeaderVarDef(
        name='$DIMTOL',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=3700,
        default=0),
    '$DIMLIM': HeaderVarDef(
        name='$DIMLIM',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=3800,
        default=0),
    '$DIMTIH': HeaderVarDef(
        name='$DIMTIH',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=3900,
        default=0),
    '$DIMTOH': HeaderVarDef(
        name='$DIMTOH',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=4000,
        default=0),
    '$DIMSE1': HeaderVarDef(
        name='$DIMSE1',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=4100,
        default=0),
    '$DIMSE2': HeaderVarDef(
        name='$DIMSE2',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=4200,
        default=0),
    '$DIMTAD': HeaderVarDef(
        name='$DIMTAD',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=4300,
        default=1),
    '$DIMZIN': HeaderVarDef(
        name='$DIMZIN',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=4400,
        default=8),
    '$DIMBLK': HeaderVarDef(
        name='$DIMBLK',
        code=1,
        factory=partial(SingleValue, code=1),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=4500,
        default=''),
    '$DIMASO': HeaderVarDef(
        name='$DIMASO',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=4600,
        default=1),
    '$DIMSHO': HeaderVarDef(
        name='$DIMSHO',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=4700,
        default=1),
    '$DIMPOST': HeaderVarDef(
        name='$DIMPOST',
        code=1,
        factory=partial(SingleValue, code=1),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=4800,
        default=''),
    '$DIMAPOST': HeaderVarDef(
        name='$DIMAPOST',
        code=1,
        factory=partial(SingleValue, code=1),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=4900,
        default=''),
    '$DIMALT': HeaderVarDef(
        name='$DIMALT',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=5000,
        default=0),
    '$DIMALTD': HeaderVarDef(
        name='$DIMALTD',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=5100,
        default=3),
    '$DIMALTF': HeaderVarDef(
        name='$DIMALTF',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=5200,
        default=0.03937007874),
    '$DIMLFAC': HeaderVarDef(
        name='$DIMLFAC',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=5300,
        default=1.0),
    '$DIMTOFL': HeaderVarDef(
        name='$DIMTOFL',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=5400,
        default=1),
    '$DIMTVP': HeaderVarDef(
        name='$DIMTVP',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=5500,
        default=0.0),
    '$DIMTIX': HeaderVarDef(
        name='$DIMTIX',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=5600,
        default=0),
    '$DIMSOXD': HeaderVarDef(
        name='$DIMSOXD',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=5700,
        default=0),
    '$DIMSAH': HeaderVarDef(
        name='$DIMSAH',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=5800,
        default=0),
    '$DIMBLK1': HeaderVarDef(
        name='$DIMBLK1',
        code=1,
        factory=partial(SingleValue, code=1),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=5900,
        default=''),
    '$DIMBLK2': HeaderVarDef(
        name='$DIMBLK2',
        code=1,
        factory=partial(SingleValue, code=1),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=6000,
        default=''),
    '$DIMSTYLE': HeaderVarDef(
        name='$DIMSTYLE',
        code=2,
        factory=partial(SingleValue, code=2),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=6100,
        default='ISO-25'),
    '$DIMCLRD': HeaderVarDef(
        name='$DIMCLRD',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=6200,
        default=0),
    '$DIMCLRE': HeaderVarDef(
        name='$DIMCLRE',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=6300,
        default=0),
    '$DIMCLRT': HeaderVarDef(
        name='$DIMCLRT',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=6400,
        default=0),
    '$DIMTFAC': HeaderVarDef(
        name='$DIMTFAC',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=6500,
        default=1.0),
    '$DIMGAP': HeaderVarDef(
        name='$DIMGAP',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=6600,
        default=0.625),
    '$DIMJUST': HeaderVarDef(
        name='$DIMJUST',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=6700,
        default=0),
    '$DIMSD1': HeaderVarDef(
        name='$DIMSD1',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=6800,
        default=0),
    '$DIMSD2': HeaderVarDef(
        name='$DIMSD2',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=6900,
        default=0),
    '$DIMTOLJ': HeaderVarDef(
        name='$DIMTOLJ',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=7000,
        default=0),
    '$DIMTZIN': HeaderVarDef(
        name='$DIMTZIN',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=7100,
        default=8),
    '$DIMALTZ': HeaderVarDef(
        name='$DIMALTZ',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=7200,
        default=0),
    '$DIMALTTZ': HeaderVarDef(
        name='$DIMALTTZ',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=7300,
        default=0),
    '$DIMUPT': HeaderVarDef(
        name='$DIMUPT',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=7400,
        default=0),
    '$DIMDEC': HeaderVarDef(
        name='$DIMDEC',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=7500,
        default=2),
    '$DIMTDEC': HeaderVarDef(
        name='$DIMTDEC',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=7600,
        default=2),
    '$DIMALTU': HeaderVarDef(
        name='$DIMALTU',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=7700,
        default=2),
    '$DIMALTTD': HeaderVarDef(
        name='$DIMALTTD',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=7800,
        default=3),
    '$DIMTXSTY': HeaderVarDef(
        name='$DIMTXSTY',
        code=7,
        factory=partial(SingleValue, code=7),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=7900,
        default='Standard'),
    '$DIMAUNIT': HeaderVarDef(
        name='$DIMAUNIT',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=8000,
        default=0),
    '$DIMADEC': HeaderVarDef(
        name='$DIMADEC',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=8100,
        default=0),
    '$DIMALTRND': HeaderVarDef(
        name='$DIMALTRND',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=8200,
        default=0.0),
    '$DIMAZIN': HeaderVarDef(
        name='$DIMAZIN',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=8300,
        default=0),
    '$DIMDSEP': HeaderVarDef(
        name='$DIMDSEP',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=8400,
        default=44),
    '$DIMATFIT': HeaderVarDef(
        name='$DIMATFIT',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=8500,
        default=3),
    '$DIMFRAC': HeaderVarDef(
        name='$DIMFRAC',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=8600,
        default=0),
    '$DIMLDRBLK': HeaderVarDef(
        name='$DIMLDRBLK',
        code=1,
        factory=partial(SingleValue, code=1),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=8700,
        default=''),
    '$DIMLUNIT': HeaderVarDef(
        name='$DIMLUNIT',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=8800,
        default=2),
    '$COORDS': HeaderVarDef(
        name='$COORDS',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF12,
        priority=8800,
        default=1),
    '$DIMLWD': HeaderVarDef(
        name='$DIMLWD',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=8900,
        default=-2),
    '$DIMLWE': HeaderVarDef(
        name='$DIMLWE',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=9000,
        default=-2),
    '$DIMTMOVE': HeaderVarDef(
        name='$DIMTMOVE',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=9100,
        default=0),
    '$DIMFXL': HeaderVarDef(
        name='$DIMFXL',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=9200,
        default=1.0),
    '$ATTDIA': HeaderVarDef(
        name='$ATTDIA',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF12,
        priority=9200,
        default=0),
    '$DIMFXLON': HeaderVarDef(
        name='$DIMFXLON',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=9300,
        default=0),
    '$ATTREQ': HeaderVarDef(
        name='$ATTREQ',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF12,
        priority=9300,
        default=1),
    '$DIMJOGANG': HeaderVarDef(
        name='$DIMJOGANG',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=9400,
        default=0.785398163397),
    '$DIMTFILL': HeaderVarDef(
        name='$DIMTFILL',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=9500,
        default=0),
    '$DIMTFILLCLR': HeaderVarDef(
        name='$DIMTFILLCLR',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=9600,
        default=0),
    '$DIMARCSYM': HeaderVarDef(
        name='$DIMARCSYM',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=9700,
        default=0),
    '$DIMLTYPE': HeaderVarDef(
        name='$DIMLTYPE',
        code=6,
        factory=partial(SingleValue, code=6),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=9800,
        default=''),
    '$DIMLTEX1': HeaderVarDef(
        name='$DIMLTEX1',
        code=6,
        factory=partial(SingleValue, code=6),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=9900,
        default=''),
    '$DIMLTEX2': HeaderVarDef(
        name='$DIMLTEX2',
        code=6,
        factory=partial(SingleValue, code=6),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=10000,
        default=''),
    '$DIMTXTDIRECTION': HeaderVarDef(
        name='$DIMTXTDIRECTION',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2010,
        maxdxf=DXF2018,
        priority=10100,
        default=0),
    '$LUNITS': HeaderVarDef(
        name='$LUNITS',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=10200,
        default=2),
    '$LUPREC': HeaderVarDef(
        name='$LUPREC',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=10300,
        default=4),
    '$SKETCHINC': HeaderVarDef(
        name='$SKETCHINC',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=10400,
        default=1.0),
    '$FILLETRAD': HeaderVarDef(
        name='$FILLETRAD',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=10500,
        default=10.0),
    '$AUNITS': HeaderVarDef(
        name='$AUNITS',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=10600,
        default=0),
    '$AUPREC': HeaderVarDef(
        name='$AUPREC',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=10700,
        default=2),
    '$MENU': HeaderVarDef(
        name='$MENU',
        code=1,
        factory=partial(SingleValue, code=1),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=10800,
        default='.'),
    '$ELEVATION': HeaderVarDef(
        name='$ELEVATION',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=10900,
        default=0.0),
    '$PELEVATION': HeaderVarDef(
        name='$PELEVATION',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=11000,
        default=0.0),
    '$THICKNESS': HeaderVarDef(
        name='$THICKNESS',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=11100,
        default=0.0),
    '$LIMCHECK': HeaderVarDef(
        name='$LIMCHECK',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=11200,
        default=0),
    '$CHAMFERA': HeaderVarDef(
        name='$CHAMFERA',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=11300,
        default=0.0),
    '$CHAMFERB': HeaderVarDef(
        name='$CHAMFERB',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=11400,
        default=0.0),
    '$CHAMFERC': HeaderVarDef(
        name='$CHAMFERC',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=11500,
        default=0.0),
    '$CHAMFERD': HeaderVarDef(
        name='$CHAMFERD',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=11600,
        default=0.0),
    '$SKPOLY': HeaderVarDef(
        name='$SKPOLY',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=11700,
        default=0),
    '$TDCREATE': HeaderVarDef(
        name='$TDCREATE',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=11800,
        default=2458532.195663565),
    '$TDUCREATE': HeaderVarDef(
        name='$TDUCREATE',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=11900,
        default=2458532.153996898),
    '$TDUPDATE': HeaderVarDef(
        name='$TDUPDATE',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=12000,
        default=2458532.196097766),
    '$TDUUPDATE': HeaderVarDef(
        name='$TDUUPDATE',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=12100,
        default=2458532.1544311),
    '$TDINDWG': HeaderVarDef(
        name='$TDINDWG',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=12200,
        default=0.0),
    '$TDUSRTIMER': HeaderVarDef(
        name='$TDUSRTIMER',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=12300,
        default=0.0),
    '$USRTIMER': HeaderVarDef(
        name='$USRTIMER',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=12400,
        default=1),
    '$ANGBASE': HeaderVarDef(
        name='$ANGBASE',
        code=50,
        factory=partial(SingleValue, code=50),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=12500,
        default=0.0),
    '$ANGDIR': HeaderVarDef(
        name='$ANGDIR',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=12600,
        default=0),
    '$PDMODE': HeaderVarDef(
        name='$PDMODE',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=12700,
        default=0),
    '$PDSIZE': HeaderVarDef(
        name='$PDSIZE',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=12800,
        default=0.0),
    '$PLINEWID': HeaderVarDef(
        name='$PLINEWID',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=12900,
        default=0.0),
    '$SPLFRAME': HeaderVarDef(
        name='$SPLFRAME',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=13000,
        default=0),
    '$SPLINETYPE': HeaderVarDef(
        name='$SPLINETYPE',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=13100,
        default=6),
    '$SPLINESEGS': HeaderVarDef(
        name='$SPLINESEGS',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=13200,
        default=8),
    '$HANDLING': HeaderVarDef(
        name='$HANDLING',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF12,
        priority=9400,
        default=1),
    '$HANDSEED': HeaderVarDef(
        name='$HANDSEED',
        code=5,
        factory=partial(SingleValue, code=5),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=13300,
        default='100'),
    '$SURFTAB1': HeaderVarDef(
        name='$SURFTAB1',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=13400,
        default=6),
    '$SURFTAB2': HeaderVarDef(
        name='$SURFTAB2',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=13500,
        default=6),
    '$SURFTYPE': HeaderVarDef(
        name='$SURFTYPE',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=13600,
        default=6),
    '$SURFU': HeaderVarDef(
        name='$SURFU',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=13700,
        default=6),
    '$SURFV': HeaderVarDef(
        name='$SURFV',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=13800,
        default=6),
    '$UCSBASE': HeaderVarDef(
        name='$UCSBASE',
        code=2,
        factory=partial(SingleValue, code=2),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=13900,
        default=''),
    '$UCSNAME': HeaderVarDef(
        name='$UCSNAME',
        code=2,
        factory=partial(SingleValue, code=2),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=14000,
        default=''),
    '$UCSORG': HeaderVarDef(
        name='$UCSORG',
        code=10,
        factory=Point3D,
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=14100,
        default=(0.0, 0.0, 0.0)),
    '$UCSXDIR': HeaderVarDef(
        name='$UCSXDIR',
        code=10,
        factory=Point3D,
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=14200,
        default=(1.0, 0.0, 0.0)),
    '$UCSYDIR': HeaderVarDef(
        name='$UCSYDIR',
        code=10,
        factory=Point3D,
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=14300,
        default=(0.0, 1.0, 0.0)),
    '$UCSORTHOREF': HeaderVarDef(
        name='$UCSORTHOREF',
        code=2,
        factory=partial(SingleValue, code=2),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=14400,
        default=''),
    '$UCSORTHOVIEW': HeaderVarDef(
        name='$UCSORTHOVIEW',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=14500,
        default=0),
    '$UCSORGTOP': HeaderVarDef(
        name='$UCSORGTOP',
        code=10,
        factory=Point3D,
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=14600,
        default=(0.0, 0.0, 0.0)),
    '$UCSORGBOTTOM': HeaderVarDef(
        name='$UCSORGBOTTOM',
        code=10,
        factory=Point3D,
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=14700,
        default=(0.0, 0.0, 0.0)),
    '$UCSORGLEFT': HeaderVarDef(
        name='$UCSORGLEFT',
        code=10,
        factory=Point3D,
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=14800,
        default=(0.0, 0.0, 0.0)),
    '$UCSORGRIGHT': HeaderVarDef(
        name='$UCSORGRIGHT',
        code=10,
        factory=Point3D,
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=14900,
        default=(0.0, 0.0, 0.0)),
    '$UCSORGFRONT': HeaderVarDef(
        name='$UCSORGFRONT',
        code=10,
        factory=Point3D,
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=15000,
        default=(0.0, 0.0, 0.0)),
    '$UCSORGBACK': HeaderVarDef(
        name='$UCSORGBACK',
        code=10,
        factory=Point3D,
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=15100,
        default=(0.0, 0.0, 0.0)),
    '$PUCSBASE': HeaderVarDef(
        name='$PUCSBASE',
        code=2,
        factory=partial(SingleValue, code=2),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=15200,
        default=''),
    '$PUCSNAME': HeaderVarDef(
        name='$PUCSNAME',
        code=2,
        factory=partial(SingleValue, code=2),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=15300,
        default=''),
    '$PUCSORG': HeaderVarDef(
        name='$PUCSORG',
        code=10,
        factory=Point3D,
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=15400,
        default=(0.0, 0.0, 0.0)),
    '$PUCSXDIR': HeaderVarDef(
        name='$PUCSXDIR',
        code=10,
        factory=Point3D,
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=15500,
        default=(1.0, 0.0, 0.0)),
    '$PUCSYDIR': HeaderVarDef(
        name='$PUCSYDIR',
        code=10,
        factory=Point3D,
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=15600,
        default=(0.0, 1.0, 0.0)),
    '$PUCSORTHOREF': HeaderVarDef(
        name='$PUCSORTHOREF',
        code=2,
        factory=partial(SingleValue, code=2),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=15700,
        default=''),
    '$PUCSORTHOVIEW': HeaderVarDef(
        name='$PUCSORTHOVIEW',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=15800,
        default=0),
    '$PUCSORGTOP': HeaderVarDef(
        name='$PUCSORGTOP',
        code=10,
        factory=Point3D,
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=15900,
        default=(0.0, 0.0, 0.0)),
    '$PUCSORGBOTTOM': HeaderVarDef(
        name='$PUCSORGBOTTOM',
        code=10,
        factory=Point3D,
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=16000,
        default=(0.0, 0.0, 0.0)),
    '$PUCSORGLEFT': HeaderVarDef(
        name='$PUCSORGLEFT',
        code=10,
        factory=Point3D,
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=16100,
        default=(0.0, 0.0, 0.0)),
    '$PUCSORGRIGHT': HeaderVarDef(
        name='$PUCSORGRIGHT',
        code=10,
        factory=Point3D,
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=16200,
        default=(0.0, 0.0, 0.0)),
    '$PUCSORGFRONT': HeaderVarDef(
        name='$PUCSORGFRONT',
        code=10,
        factory=Point3D,
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=16300,
        default=(0.0, 0.0, 0.0)),
    '$PUCSORGBACK': HeaderVarDef(
        name='$PUCSORGBACK',
        code=10,
        factory=Point3D,
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=16400,
        default=(0.0, 0.0, 0.0)),
    '$USERI1': HeaderVarDef(
        name='$USERI1',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=16500,
        default=0),
    '$USERI2': HeaderVarDef(
        name='$USERI2',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=16600,
        default=0),
    '$USERI3': HeaderVarDef(
        name='$USERI3',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=16700,
        default=0),
    '$USERI4': HeaderVarDef(
        name='$USERI4',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=16800,
        default=0),
    '$USERI5': HeaderVarDef(
        name='$USERI5',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=16900,
        default=0),
    '$USERR1': HeaderVarDef(
        name='$USERR1',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=17000,
        default=0.0),
    '$USERR2': HeaderVarDef(
        name='$USERR2',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=17100,
        default=0.0),
    '$USERR3': HeaderVarDef(
        name='$USERR3',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=17200,
        default=0.0),
    '$USERR4': HeaderVarDef(
        name='$USERR4',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=17300,
        default=0.0),
    '$USERR5': HeaderVarDef(
        name='$USERR5',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=17400,
        default=0.0),
    '$WORLDVIEW': HeaderVarDef(
        name='$WORLDVIEW',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=17500,
        default=1),
    '$SHADEDGE': HeaderVarDef(
        name='$SHADEDGE',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=17600,
        default=3),
    '$SHADEDIF': HeaderVarDef(
        name='$SHADEDIF',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=17700,
        default=70),
    '$TILEMODE': HeaderVarDef(
        name='$TILEMODE',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=17800,
        default=1),
    '$MAXACTVP': HeaderVarDef(
        name='$MAXACTVP',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=17900,
        default=64),
    '$PINSBASE': HeaderVarDef(
        name='$PINSBASE',
        code=10,
        factory=Point3D,
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=18000,
        default=(0.0, 0.0, 0.0)),
    '$PLIMCHECK': HeaderVarDef(
        name='$PLIMCHECK',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=18100,
        default=0),
    '$PEXTMIN': HeaderVarDef(
        name='$PEXTMIN',
        code=10,
        factory=Point3D,
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=18200,
        default=(1e+20, 1e+20, 1e+20)),
    '$PEXTMAX': HeaderVarDef(
        name='$PEXTMAX',
        code=10,
        factory=Point3D,
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=18300,
        default=(-1e+20, -1e+20, -1e+20)),
    '$PLIMMIN': HeaderVarDef(
        name='$PLIMMIN',
        code=10,
        factory=Point2D,
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=18400,
        default=(0.0, 0.0)),
    '$PLIMMAX': HeaderVarDef(
        name='$PLIMMAX',
        code=10,
        factory=Point2D,
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=18500,
        default=(420.0, 297.0)),
    '$UNITMODE': HeaderVarDef(
        name='$UNITMODE',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=18600,
        default=0),
    '$VISRETAIN': HeaderVarDef(
        name='$VISRETAIN',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=18700,
        default=1),
    '$PLINEGEN': HeaderVarDef(
        name='$PLINEGEN',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=18800,
        default=0),
    '$PSLTSCALE': HeaderVarDef(
        name='$PSLTSCALE',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=18900,
        default=1),
    '$TREEDEPTH': HeaderVarDef(
        name='$TREEDEPTH',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=19000,
        default=3020),
    '$CMLSTYLE': HeaderVarDef(
        name='$CMLSTYLE',
        code=2,
        factory=partial(SingleValue, code=2),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=19100,
        default='Standard'),
    '$CMLJUST': HeaderVarDef(
        name='$CMLJUST',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=19200,
        default=0),
    '$CMLSCALE': HeaderVarDef(
        name='$CMLSCALE',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=19300,
        default=20.0),
    '$PROXYGRAPHICS': HeaderVarDef(
        name='$PROXYGRAPHICS',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=19400,
        default=1),
    '$MEASUREMENT': HeaderVarDef(
        name='$MEASUREMENT',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=19500,
        default=1),
    '$CELWEIGHT': HeaderVarDef(
        name='$CELWEIGHT',
        code=370,
        factory=partial(SingleValue, code=370),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=19600,
        default=-1),
    '$ENDCAPS': HeaderVarDef(
        name='$ENDCAPS',
        code=280,
        factory=partial(SingleValue, code=280),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=19700,
        default=0),
    '$JOINSTYLE': HeaderVarDef(
        name='$JOINSTYLE',
        code=280,
        factory=partial(SingleValue, code=280),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=19800,
        default=0),
    '$LWDISPLAY': HeaderVarDef(
        name='$LWDISPLAY',
        code=290,
        factory=partial(SingleValue, code=290),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=19900,
        default=0),
    '$INSUNITS': HeaderVarDef(
        name='$INSUNITS',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=20000,
        default=6),
    '$HYPERLINKBASE': HeaderVarDef(
        name='$HYPERLINKBASE',
        code=1,
        factory=partial(SingleValue, code=1),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=20100,
        default=''),
    '$STYLESHEET': HeaderVarDef(
        name='$STYLESHEET',
        code=1,
        factory=partial(SingleValue, code=1),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=20200,
        default=''),
    '$XEDIT': HeaderVarDef(
        name='$XEDIT',
        code=290,
        factory=partial(SingleValue, code=290),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=20300,
        default=1),
    '$CEPSNTYPE': HeaderVarDef(
        name='$CEPSNTYPE',
        code=380,
        factory=partial(SingleValue, code=380),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=20400,
        default=0),
    '$PSTYLEMODE': HeaderVarDef(
        name='$PSTYLEMODE',
        code=290,
        factory=partial(SingleValue, code=290),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=20500,
        default=1),
    '$FINGERPRINTGUID': HeaderVarDef(
        name='$FINGERPRINTGUID',
        code=2,
        factory=partial(SingleValue, code=2),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=20600,
        default='{6d50f5e5-c3bb-4eb1-8520-1109eb28619f}'),
    '$VERSIONGUID': HeaderVarDef(
        name='$VERSIONGUID',
        code=2,
        factory=partial(SingleValue, code=2),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=20700,
        default='{FAEB1C32-E019-11D5-929B-00C0DF256EC4}'),
    '$EXTNAMES': HeaderVarDef(
        name='$EXTNAMES',
        code=290,
        factory=partial(SingleValue, code=290),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=20800,
        default=1),
    '$PSVPSCALE': HeaderVarDef(
        name='$PSVPSCALE',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=20900,
        default=0.0),
    '$OLESTARTUP': HeaderVarDef(
        name='$OLESTARTUP',
        code=290,
        factory=partial(SingleValue, code=290),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=21000,
        default=0),
    '$SORTENTS': HeaderVarDef(
        name='$SORTENTS',
        code=280,
        factory=partial(SingleValue, code=280),
        mindxf=DXF2004,
        maxdxf=DXF2018,
        priority=21100,
        default=127),
    '$INDEXCTL': HeaderVarDef(
        name='$INDEXCTL',
        code=280,
        factory=partial(SingleValue, code=280),
        mindxf=DXF2004,
        maxdxf=DXF2018,
        priority=21200,
        default=0),
    '$HIDETEXT': HeaderVarDef(
        name='$HIDETEXT',
        code=280,
        factory=partial(SingleValue, code=280),
        mindxf=DXF2004,
        maxdxf=DXF2018,
        priority=21300,
        default=1),
    '$XCLIPFRAME': HeaderVarDef(
        name='$XCLIPFRAME',
        code=280,
        factory=partial(SingleValue, code=280),
        mindxf=DXF2004,
        maxdxf=DXF2018,
        priority=21400,
        default=2),
    '$HALOGAP': HeaderVarDef(
        name='$HALOGAP',
        code=280,
        factory=partial(SingleValue, code=280),
        mindxf=DXF2004,
        maxdxf=DXF2018,
        priority=21500,
        default=0),
    '$OBSCOLOR': HeaderVarDef(
        name='$OBSCOLOR',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2004,
        maxdxf=DXF2018,
        priority=21600,
        default=257),
    '$OBSLTYPE': HeaderVarDef(
        name='$OBSLTYPE',
        code=280,
        factory=partial(SingleValue, code=280),
        mindxf=DXF2004,
        maxdxf=DXF2018,
        priority=21700,
        default=0),
    '$INTERSECTIONDISPLAY': HeaderVarDef(
        name='$INTERSECTIONDISPLAY',
        code=280,
        factory=partial(SingleValue, code=280),
        mindxf=DXF2004,
        maxdxf=DXF2018,
        priority=21800,
        default=0),
    '$INTERSECTIONCOLOR': HeaderVarDef(
        name='$INTERSECTIONCOLOR',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2004,
        maxdxf=DXF2018,
        priority=21900,
        default=257),
    '$DIMASSOC': HeaderVarDef(
        name='$DIMASSOC',
        code=280,
        factory=partial(SingleValue, code=280),
        mindxf=DXF2004,
        maxdxf=DXF2018,
        priority=22000,
        default=2),
    '$PROJECTNAME': HeaderVarDef(
        name='$PROJECTNAME',
        code=1,
        factory=partial(SingleValue, code=1),
        mindxf=DXF2004,
        maxdxf=DXF2018,
        priority=22100,
        default=''),
    '$CAMERADISPLAY': HeaderVarDef(
        name='$CAMERADISPLAY',
        code=290,
        factory=partial(SingleValue, code=290),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=22200,
        default=0),
    '$LENSLENGTH': HeaderVarDef(
        name='$LENSLENGTH',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=22300,
        default=50.0),
    '$CAMERAHEIGHT': HeaderVarDef(
        name='$CAMERAHEIGHT',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=22400,
        default=0.0),
    '$STEPSPERSEC': HeaderVarDef(
        name='$STEPSPERSEC',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=22500,
        default=24.0),
    '$STEPSIZE': HeaderVarDef(
        name='$STEPSIZE',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=22600,
        default=100.0),
    '$3DDWFPREC': HeaderVarDef(
        name='$3DDWFPREC',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=22700,
        default=2.0),
    '$PSOLWIDTH': HeaderVarDef(
        name='$PSOLWIDTH',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=22800,
        default=0.005),
    '$PSOLHEIGHT': HeaderVarDef(
        name='$PSOLHEIGHT',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=22900,
        default=0.08),
    '$LOFTANG1': HeaderVarDef(
        name='$LOFTANG1',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=23000,
        default=1.570796326795),
    '$LOFTANG2': HeaderVarDef(
        name='$LOFTANG2',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=23100,
        default=1.570796326795),
    '$LOFTMAG1': HeaderVarDef(
        name='$LOFTMAG1',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=23200,
        default=0.0),
    '$LOFTMAG2': HeaderVarDef(
        name='$LOFTMAG2',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=23300,
        default=0.0),
    '$LOFTPARAM': HeaderVarDef(
        name='$LOFTPARAM',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=23400,
        default=7),
    '$LOFTNORMALS': HeaderVarDef(
        name='$LOFTNORMALS',
        code=280,
        factory=partial(SingleValue, code=280),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=23500,
        default=1),
    '$LATITUDE': HeaderVarDef(
        name='$LATITUDE',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=23600,
        default=37.795),
    '$LONGITUDE': HeaderVarDef(
        name='$LONGITUDE',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=23700,
        default=-122.394),
    '$NORTHDIRECTION': HeaderVarDef(
        name='$NORTHDIRECTION',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=23800,
        default=0.0),
    '$TIMEZONE': HeaderVarDef(
        name='$TIMEZONE',
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=23900,
        default=-8000),
    '$LIGHTGLYPHDISPLAY': HeaderVarDef(
        name='$LIGHTGLYPHDISPLAY',
        code=280,
        factory=partial(SingleValue, code=280),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=24000,
        default=1),
    '$TILEMODELIGHTSYNCH': HeaderVarDef(
        name='$TILEMODELIGHTSYNCH',
        code=280,
        factory=partial(SingleValue, code=280),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=24100,
        default=1),
    '$CMATERIAL': HeaderVarDef(
        name='$CMATERIAL',
        code=347,
        factory=partial(SingleValue, code=347),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=24200,
        default='45'),  # default value '0' crashes BricsCAD
    '$SOLIDHIST': HeaderVarDef(
        name='$SOLIDHIST',
        code=280,
        factory=partial(SingleValue, code=280),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=24300,
        default=0),
    '$SHOWHIST': HeaderVarDef(
        name='$SHOWHIST',
        code=280,
        factory=partial(SingleValue, code=280),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=24400,
        default=1),
    '$DWFFRAME': HeaderVarDef(
        name='$DWFFRAME',
        code=280,
        factory=partial(SingleValue, code=280),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=24500,
        default=2),
    '$DGNFRAME': HeaderVarDef(
        name='$DGNFRAME',
        code=280,
        factory=partial(SingleValue, code=280),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=24600,
        default=2),
    '$REALWORLDSCALE': HeaderVarDef(
        name='$REALWORLDSCALE',
        code=290,
        factory=partial(SingleValue, code=290),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=24700,
        default=1),
    '$INTERFERECOLOR': HeaderVarDef(
        name='$INTERFERECOLOR',
        code=62,
        factory=partial(SingleValue, code=62),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=24800,
        default=256),
    '$INTERFEREOBJVS': HeaderVarDef(
        # Handle to a VisualStyle, if pointing to an invalid or non-existing VS,
        # copy to clipboard in AutoCAD can fail.
        name='$INTERFEREOBJVS',
        code=345,
        factory=partial(SingleValue, code=345),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=24810,
        default=None),  # will not be written, if not set
    '$INTERFEREVPVS': HeaderVarDef(
        # Handle to a VisualStyle, if pointing to an invalid or non-existing VS,
        # copy to clipboard in AutoCAD can fail.
        name='$INTERFEREVPVS',
        code=346,
        factory=partial(SingleValue, code=346),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=24820,
        default=None),  # will not be written, if not set
    '$CSHADOW': HeaderVarDef(
        name='$CSHADOW',
        code=280,
        factory=partial(SingleValue, code=280),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=24900,
        default=0),
    '$SHADOWPLANELOCATION': HeaderVarDef(
        name='$SHADOWPLANELOCATION',
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=25000,
        default=0.0),
}



# Copyright (c) 2019-2020 Manfred Moitzi
# License: MIT License
from typing import TYPE_CHECKING
from ezdxf.lldxf import validator
from ezdxf.lldxf.attributes import (
    DXFAttr, DXFAttributes, DefSubclass, XType, RETURN_DEFAULT,
    group_code_mapping,
)
from ezdxf.lldxf.const import SUBCLASS_MARKER, DXF2000
from ezdxf.math import Vec3, Matrix44, NULLVEC, Z_AXIS
from .dxfentity import base_class, SubclassProcessor
from .dxfgfx import DXFGraphic, acdb_entity
from .factory import register_entity

if TYPE_CHECKING:
    from ezdxf.eztypes import TagWriter, DXFNamespace

__all__ = ['Ray', 'XLine']

acdb_xline = DefSubclass('AcDbXline', {
    'start': DXFAttr(10, xtype=XType.point3d, default=NULLVEC),
    'unit_vector': DXFAttr(
        11, xtype=XType.point3d, default=Z_AXIS,
        validator=validator.is_not_null_vector,
        fixer=RETURN_DEFAULT,
    ),
})
acdb_xline_group_codes = group_code_mapping(acdb_xline)


@register_entity
class XLine(DXFGraphic):
    """ DXF XLINE entity """
    DXFTYPE = 'XLINE'
    DXFATTRIBS = DXFAttributes(base_class, acdb_entity, acdb_xline)
    MIN_DXF_VERSION_FOR_EXPORT = DXF2000
    XLINE_SUBCLASS = 'AcDbXline'

    def load_dxf_attribs(
            self, processor: SubclassProcessor = None) -> 'DXFNamespace':
        dxf = super().load_dxf_attribs(processor)
        if processor:
            processor.fast_load_dxfattribs(
                dxf, acdb_xline_group_codes, subclass=2, recover=True)
        return dxf

    def export_entity(self, tagwriter: 'TagWriter') -> None:
        """ Export entity specific data as DXF tags. """
        super().export_entity(tagwriter)
        tagwriter.write_tag2(SUBCLASS_MARKER, self.XLINE_SUBCLASS)
        self.dxf.export_dxf_attribs(tagwriter, ['start', 'unit_vector'])

    def transform(self, m: Matrix44) -> 'XLine':
        """ Transform XLINE/RAY entity by transformation matrix `m` inplace.

        .. versionadded:: 0.13

        """
        self.dxf.start = m.transform(self.dxf.start)
        self.dxf.unit_vector = m.transform_direction(
            self.dxf.unit_vector).normalize()
        return self

    def translate(self, dx: float, dy: float, dz: float) -> 'XLine':
        """ Optimized XLINE/RAY translation about `dx` in x-axis, `dy` in
        y-axis and `dz` in z-axis, returns `self` (floating interface).

        .. versionadded:: 0.13

        """
        self.dxf.start = Vec3(dx, dy, dz) + self.dxf.start
        return self


@register_entity
class Ray(XLine):
    """ DXF Ray entity """
    DXFTYPE = 'RAY'
    DXFATTRIBS = DXFAttributes(base_class, acdb_entity, acdb_xline)
    MIN_DXF_VERSION_FOR_EXPORT = DXF2000
    XLINE_SUBCLASS = 'AcDbRay'



# Copyright (c) 2014-2020, Manfred Moitzi
# License: MIT License
from typing import BinaryIO, cast, TextIO, List, Optional
import zipfile
from contextlib import contextmanager

from ezdxf.lldxf.validator import is_dxf_stream, dxf_info

CRLF = b'\r\n'
LF = b'\n'


class ZipReader:
    def __init__(self, zip_archive_name: str, errors='surrogateescape'):
        if not zipfile.is_zipfile(zip_archive_name):
            raise IOError(f"'{zip_archive_name}' is not a zip archive.")
        self.zip_archive_name = zip_archive_name
        self.zip_archive: Optional[zipfile.ZipFile] = None
        self.dxf_file_name: Optional[str] = None
        self.dxf_file: Optional[BinaryIO] = None
        self.encoding = 'cp1252'
        self.errors = errors
        self.dxfversion = 'AC1009'

    def open(self, dxf_file_name: str = None) -> None:
        def open_dxf_file() -> BinaryIO:
            # Open always in binary mode:
            return cast(BinaryIO, self.zip_archive.open(self.dxf_file_name))

        self.zip_archive = zipfile.ZipFile(self.zip_archive_name)
        self.dxf_file_name = dxf_file_name if dxf_file_name is not None \
            else self.get_first_dxf_file_name()
        self.dxf_file = open_dxf_file()

        # Reading with standard encoding 'cp1252' - readline() fails if leading
        # comments contain none ASCII characters.
        if not is_dxf_stream(cast(TextIO, self)):
            raise IOError(f"'{self.dxf_file_name}' is not a DXF file.")
        self.dxf_file = open_dxf_file()  # restart
        self.get_dxf_info()
        self.dxf_file = open_dxf_file()  # restart

    def get_first_dxf_file_name(self) -> str:
        dxf_file_names = self.get_dxf_file_names()
        if len(dxf_file_names) > 0:
            return dxf_file_names[0]
        else:
            raise IOError("No DXF files found.")

    def get_dxf_file_names(self) -> List[str]:
        return [name for name in self.zip_archive.namelist()
                if name.lower().endswith('.dxf')]

    def get_dxf_info(self) -> None:
        info = dxf_info(cast(TextIO, self))
        # Since DXF R2007 (AC1021) file encoding is always 'utf-8'
        self.encoding = info.encoding if info.version < 'AC1021' else 'utf-8'
        self.dxfversion = info.version

    # Required TextIO interface
    def readline(self) -> str:
        next_line = self.dxf_file.readline().replace(CRLF, LF)
        return str(next_line, self.encoding, self.errors)

    def close(self) -> None:
        self.zip_archive.close()


@contextmanager
def ctxZipReader(zipfilename: str, filename: str = None,
                 errors: str = 'surrogateescape') -> ZipReader:
    zip_reader = ZipReader(zipfilename, errors=errors)
    zip_reader.open(filename)
    yield zip_reader
    zip_reader.close()



# Created: 13.03.2011
# Copyright (c) 2011-2019, Manfred Moitzi
# License: MIT License
from typing import TYPE_CHECKING, Iterable, Tuple, cast, Iterator, Union
import logging

from ezdxf.entities.dictionary import Dictionary
from ezdxf.entities import factory
from ezdxf.lldxf.const import DXFStructureError, DXFValueError, RASTER_UNITS, DXFKeyError
from ezdxf.entitydb import EntitySpace
from ezdxf.query import EntityQuery
from ezdxf.tools.handle import UnderlayKeyGenerator

if TYPE_CHECKING:
    from ezdxf.eztypes import GeoData, DictionaryVar
    from ezdxf.eztypes import Drawing, TagWriter, EntityDB, DXFTagStorage, DXFObject
    from ezdxf.eztypes import ImageDefReactor, ImageDef, UnderlayDef, DictionaryWithDefault, XRecord, Placeholder

logger = logging.getLogger('ezdxf')


class ObjectsSection:
    def __init__(self, doc: 'Drawing', entities: Iterable['DXFObject'] = None):
        self.doc = doc
        self.underlay_key_generator = UnderlayKeyGenerator()
        self._entity_space = EntitySpace()
        if entities is not None:
            self._build(iter(entities))

    @property
    def entitydb(self) -> 'EntityDB':
        """ Returns drawing entity database. (internal API)"""
        return self.doc.entitydb

    def get_entity_space(self) -> 'EntitySpace':
        """ Returns entity space. (internal API) """
        return self._entity_space

    def next_underlay_key(self, checkfunc=lambda k: True) -> str:
        while True:
            key = self.underlay_key_generator.next()
            if checkfunc(key):
                return key

    def _build(self, entities: Iterator['DXFObject']) -> None:
        section_head = next(entities)  # type: DXFTagStorage

        if section_head.dxftype() != 'SECTION' or section_head.base_class[1] != (2, 'OBJECTS'):
            raise DXFStructureError("Critical structure error in 'OBJECTS' section.")

        for entity in entities:
            self._entity_space.add(entity)

    def export_dxf(self, tagwriter: 'TagWriter') -> None:
        """ Export DXF entity by `tagwriter`. (internal API) """
        tagwriter.write_str("  0\nSECTION\n  2\nOBJECTS\n")
        self._entity_space.export_dxf(tagwriter)
        tagwriter.write_tag2(0, "ENDSEC")

    def new_entity(self, _type: str, dxfattribs: dict) -> 'DXFObject':
        """ Create new DXF object, add it to the entity database and to the entity space.

        Args:
             _type: DXF type like `DICTIONARY`
             dxfattribs: DXF attributes as dict

        (internal API)
        """
        dxf_entity = factory.create_db_entry(_type, dxfattribs, self.doc)
        self._entity_space.add(dxf_entity)
        return dxf_entity

    def delete_entity(self, entity: 'DXFObject') -> None:
        """ Remove `entity` from entity space and destroy object. (internal API) """
        self._entity_space.remove(entity)
        self.entitydb.delete_entity(entity)

    def delete_all_entities(self) -> None:
        """ Delete all DXF objects. (internal API) """
        db = self.entitydb
        for entity in self._entity_space:
            db.delete_entity(entity)
        self._entity_space.clear()

    def setup_rootdict(self) -> Dictionary:
        """ Create a root dictionary. Has to be the first object in the objects section. (internal API) """
        if len(self):
            raise DXFStructureError("Can not create root dictionary in none empty objects section.")
        logger.debug('Creating ROOT dictionary.')
        # root directory has no owner
        return self.add_dictionary(owner='0')

    def setup_objects_management_tables(self, rootdict: Dictionary) -> None:
        """ Setup required management tables. (internal API)"""

        def setup_plot_style_name_table():
            plot_style_name_dict = self.add_dictionary_with_default(owner=rootdict.dxf.handle)
            placeholder = self.add_placeholder(owner=plot_style_name_dict.dxf.handle)
            plot_style_name_dict.set_default(placeholder)
            plot_style_name_dict['Normal'] = placeholder
            rootdict['ACAD_PLOTSTYLENAME'] = plot_style_name_dict

        for name in _OBJECT_TABLE_NAMES:
            if name in rootdict:
                continue  # just create not existing tables
            logger.info('creating {} dictionary'.format(name))
            if name == "ACAD_PLOTSTYLENAME":
                setup_plot_style_name_table()
            else:
                rootdict.add_new_dict(name)

    def add_object(self, entity: 'DXFObject') -> None:
        """ Add `entity` to OBJECTS section. (internal API) """
        self._entity_space.add(entity)

    def add_dxf_object_with_reactor(self, dxftype: str, dxfattribs: dict) -> 'DXFObject':
        """ Add DXF object with reactor. (internal API) """
        dxfobject = self.new_entity(dxftype, dxfattribs)
        dxfobject.set_reactors([dxfattribs['owner']])
        return dxfobject

    def purge(self):
        self._entity_space.purge()

    # start of public interface

    @property
    def rootdict(self) -> Dictionary:
        """ Root dictionary. """
        if len(self):
            return self._entity_space[0]
        else:
            return self.setup_rootdict()

    def __len__(self) -> int:
        """ Returns count of DXF objects. """
        return len(self._entity_space)

    def __iter__(self) -> Iterable['DXFObject']:
        """ Returns iterable of all DXF objects in the OBJECTS section. """
        return iter(self._entity_space)

    def __getitem__(self, index) -> 'DXFObject':
        """ Get entity at `index`.

        The underlying data structure for storing DXF objects is organized like a standard Python list, therefore `index`
        can be any valid list indexing or slicing term, like a single index ``objects[-1]`` to get the last entity, or
        an index slice ``objects[:10]`` to get the first 10 or less objects as ``List[DXFObject]``.

        """
        return self._entity_space[index]  # type: DXFObject

    def __contains__(self, entity: Union['DXFObject', str]) -> bool:
        """ Returns ``True`` if `entity` stored in OBJECTS section.

        Args:
             entity: :class:`DXFObject` or handle as hex string

        """
        if isinstance(entity, str):
            try:
                entity = self.entitydb[entity]
            except KeyError:
                return False
        return entity in self._entity_space

    def query(self, query: str = '*') -> EntityQuery:
        """
        Get all DXF objects matching the :ref:`entity query string`.

        """
        return EntityQuery(iter(self), query)

    def add_dictionary(self, owner: str = '0', hard_owned: bool = False) -> Dictionary:
        """ Add new :class:`~ezdxf.entities.Dictionary` object.

        Args:
            owner: handle to owner as hex string.
            hard_owned: ``True`` to treat entries as hard owned.

        """
        entity = self.new_entity('DICTIONARY', dxfattribs={
            'owner': owner,
            'hard_owned': hard_owned,
        })
        return cast(Dictionary, entity)

    def add_dictionary_with_default(self, owner='0', default='0', hard_owned: bool = False) -> 'DictionaryWithDefault':
        """ Add new :class:`~ezdxf.entities.DictionaryWithDefault` object.

        Args:
            owner: handle to owner as hex string.
            default: handle to default entry.
            hard_owned: ``True`` to treat entries as hard owned.

        """
        entity = self.new_entity('ACDBDICTIONARYWDFLT', dxfattribs={
            'owner': owner,
            'default': default,
            'hard_owned': hard_owned,
        })
        return cast('DictionaryWithDefault', entity)

    def add_dictionary_var(self, owner: str = '0', value: str = '') -> 'DictionaryVar':
        """
        Add a new :class:`~ezdxf.entities.DictionaryVar` object.

        Args:
            owner: handle to owner as hex string.
            value: value as string

        """
        return self.new_entity('DICTIONARYVAR', dxfattribs={'owner': owner, 'value': value})

    def add_xrecord(self, owner: str = '0') -> 'XRecord':
        """
        Add a new :class:`~ezdxf.entities.XRecord` object.

        Args:
            owner: handle to owner as hex string.

        """
        return self.new_entity('XRECORD', dxfattribs={'owner': owner})

    def add_placeholder(self, owner: str = '0') -> 'Placeholder':
        """
        Add a new :class:`~ezdxf.entities.Placeholder` object.

        Args:
            owner: handle to owner as hex string.

        """
        return self.new_entity('ACDBPLACEHOLDER', dxfattribs={'owner': owner})

    # end of public interface

    def set_raster_variables(self, frame: int = 0, quality: int = 1, units: str = 'm') -> None:
        """
        Set raster variables.

        Args:
            frame: ``0`` = do not show image frame; ``1`` = show image frame
            quality: ``0`` = draft; ``1`` = high
            units: units for inserting images. This defines the real world unit for one drawing unit for the purpose of
                   inserting and scaling images with an associated resolution.

                   ===== ===========================
                   mm    Millimeter
                   cm    Centimeter
                   m     Meter (ezdxf default)
                   km    Kilometer
                   in    Inch
                   ft    Foot
                   yd    Yard
                   mi    Mile
                   ===== ===========================

        (internal API), public interface :meth:`~ezdxf.drawing.Drawing.set_raster_variables`

        """
        units = RASTER_UNITS.get(units, 0)
        try:
            raster_vars = self.rootdict['ACAD_IMAGE_VARS']
        except DXFKeyError:
            raster_vars = self.add_dxf_object_with_reactor('RASTERVARIABLES', dxfattribs={
                'owner': self.rootdict.dxf.handle,
                'frame': frame,
                'quality': quality,
                'units': units,
            })
            self.rootdict['ACAD_IMAGE_VARS'] = raster_vars
        else:
            raster_vars.dxf.frame = frame
            raster_vars.dxf.quality = quality
            raster_vars.dxf.units = units

    def set_wipeout_variables(self, frame: int = 0) -> None:
        """
        Set wipeout variables.

        Args:
            frame: ``0`` = do not show image frame; ``1`` = show image frame

        (internal API), public interface :meth:`~ezdxf.drawing.Drawing.set_wipeout_variables`

        """
        try:
            wipeout_vars = self.rootdict['ACAD_WIPEOUT_VARS']
        except DXFKeyError:
            wipeout_vars = self.add_dxf_object_with_reactor('WIPEOUTVARIABLES', dxfattribs={
                'owner': self.rootdict.dxf.handle,
                'frame': int(frame),
            })
            self.rootdict['ACAD_WIPEOUT_VARS'] = wipeout_vars
        else:
            wipeout_vars.dxf.frame = int(frame)

    def add_image_def(self, filename: str, size_in_pixel: Tuple[int, int], name=None) -> 'ImageDef':
        """
        Add an image definition to the objects section.

        Add an :class:`~ezdxf.entities.image.ImageDef` entity to the drawing (objects section). `filename` is the image
        file name as relative or absolute path and `size_in_pixel` is the image size in pixel as (x, y) tuple. To avoid
        dependencies to external packages, `ezdxf` can not determine the image size by itself. Returns a
        :class:`~ezdxf.entities.image.ImageDef` entity which is needed to create an image reference. `name` is the
        internal image name, if set to ``None``, name is auto-generated.

        Absolute image paths works best for AutoCAD but not really good, you have to update external references manually
        in AutoCAD, which is not possible in TrueView. If the drawing units differ from 1 meter, you also have to use:
        :meth:`set_raster_variables`.

        Args:
            filename: image file name (absolute path works best for AutoCAD)
            size_in_pixel: image size in pixel as (x, y) tuple
            name: image name for internal use, None for using filename as name (best for AutoCAD)

        """
        # removed auto-generated name
        # use absolute image paths for filename and AutoCAD loads images automatically
        if name is None:
            name = filename
        image_dict = self.rootdict.get_required_dict('ACAD_IMAGE_DICT')
        image_def = self.add_dxf_object_with_reactor('IMAGEDEF', dxfattribs={
            'owner': image_dict.dxf.handle,
            'filename': filename,
            'image_size': size_in_pixel,
        })
        image_dict[name] = image_def.dxf.handle
        return cast('ImageDef', image_def)

    def add_image_def_reactor(self, image_handle: str) -> 'ImageDefReactor':
        """ Add required IMAGEDEF_REACTOR object for IMAGEDEF object. (internal API) """
        image_def_reactor = self.new_entity('IMAGEDEF_REACTOR', dxfattribs={
            'owner': image_handle,
            'image_handle': image_handle,
        })
        return cast('ImageDefReactor', image_def_reactor)

    def add_underlay_def(self, filename: str, format: str = 'pdf', name: str = None) -> 'UnderlayDef':
        """
        Add an :class:`~ezdxf.entities.underlay.UnderlayDef` entity to the drawing (OBJECTS section).
        `filename` is the underlay file name as relative or absolute path and `format` as string (pdf, dwf, dgn).
        The underlay definition is required to create an underlay reference.

        Args:
            filename: underlay file name
            format: file format as string ``'pdf'|'dwf'|'dgn'`` or ``'ext'`` for getting file format from filename extension
            name: pdf format = page number to display; dgn format = ``'default'``; dwf: ????

        """
        fmt = format.upper()
        if fmt in ('PDF', 'DWF', 'DGN'):
            underlay_dict_name = 'ACAD_{}DEFINITIONS'.format(fmt)
            underlay_def_entity = "{}DEFINITION".format(fmt)
        else:
            raise DXFValueError("Unsupported file format: '{}'".format(fmt))

        if name is None:
            if fmt == 'PDF':
                name = '1'  # Display first page by default
            elif fmt == 'DGN':
                name = 'default'
            else:
                name = 'Model'  # Display model space for DWF ???

        underlay_dict = self.rootdict.get_required_dict(underlay_dict_name)
        underlay_def = self.new_entity(underlay_def_entity, dxfattribs={
            'owner': underlay_dict.dxf.handle,
            'filename': filename,
            'name': name,
        })

        # auto-generated underlay key
        key = self.next_underlay_key(lambda k: k not in underlay_dict)
        underlay_dict[key] = underlay_def.dxf.handle
        return cast('UnderlayDef', underlay_def)

    def add_geodata(self, owner: str = '0', dxfattribs: dict = None) -> 'GeoData':
        """
        Creates a new :class:`GeoData` entity and replaces existing ones. The GEODATA entity resides in the OBJECTS section
        and NOT in the layout entity space and it is linked to the layout by an extension dictionary located in BLOCK_RECORD
        of the layout.

        The GEODATA entity requires DXF version R2010+. The DXF Reference does not document if other layouts than model
        space supports geo referencing, so getting/setting geo data may only make sense for the model space layout, but
        it is also available in paper space layouts.

        Args:
            owner: handle to owner as hex string
            dxfattribs: DXF attributes for :class:`~ezdxf.entities.GeoData` entity


        """
        if dxfattribs is None:
            dxfattribs = {}
        dxfattribs['owner'] = owner
        return cast('GeoData', self.add_dxf_object_with_reactor('GEODATA', dxfattribs))


_OBJECT_TABLE_NAMES = [
    "ACAD_COLOR",
    "ACAD_GROUP",
    "ACAD_LAYOUT",
    "ACAD_MATERIAL",
    "ACAD_MLEADERSTYLE",
    "ACAD_MLINESTYLE",
    "ACAD_PLOTSETTINGS",
    "ACAD_PLOTSTYLENAME",
    "ACAD_SCALELIST",
    "ACAD_TABLESTYLE",
    "ACAD_VISUALSTYLE",
]



# Copyright (c) 2019-2020 Manfred Moitzi
# License: MIT License
from typing import TYPE_CHECKING, Union
from ezdxf.lldxf.tags import Tags
from ezdxf.lldxf.const import DXFStructureError
from ezdxf.lldxf.const import (
    ACAD_XDICTIONARY, XDICT_HANDLE_CODE, APP_DATA_MARKER,
)

if TYPE_CHECKING:
    from ezdxf.lldxf.tagwriter import TagWriter
    from ezdxf.eztypes import Dictionary, Drawing, DXFEntity

__all__ = ['ExtensionDict']


# Example for table head and -entries with extension dicts:
# AutodeskSamples\lineweights.dxf

class ExtensionDict:
    """ Stores extended data of entities in app data 'ACAD_XDICTIONARY', app
    data contains just one entry to a hard-owned DICTIONARY objects, which is
    not shared with other entities, each entity copy has its own extension
    dictionary and the extension dictionary is destroyed when the owner entity
    is deleted from database.

    """
    __slots__ = ('_xdict',)

    def __init__(self, xdict: Union[str, 'Dictionary']):
        # 1st loading stage: xdict as string -> handle to dict
        # 2nd loading stage: xdict as DXF Dictionary
        self._xdict = xdict

    @property
    def dictionary(self) -> 'Dictionary':
        """ Get associated extension dictionary as :class:`Dictionary` object.
        """
        assert self._xdict is not None
        return self._xdict

    def __getitem__(self, key: str):
        return self.dictionary[key]

    def __setitem__(self, key: str, value):
        self.dictionary[key] = value

    def __contains__(self, key: str):
        return key in self.dictionary

    @classmethod
    def new(cls, owner_handle: str, doc: 'Drawing'):
        xdict = doc.objects.add_dictionary(
            owner=owner_handle,
            # All data in the extension dictionary belongs only to the owner
            hard_owned=True,
        )
        return cls(xdict)

    @property
    def is_alive(self):
        # Can not check if _xdict (as handle or Dictionary) really exist:
        return self._xdict is not None

    def update_owner(self, handle: str) -> None:
        """ Update owner tag of contained DXF Dictionary. """
        self.dictionary.dxf.owner = handle

    @classmethod
    def from_tags(cls, tags: Tags):
        assert tags is not None
        # Expected DXF structure:
        # [(102, '{ACAD_XDICTIONARY', (360, handle), (102, '}')]
        if len(tags) != 3 or tags[1].code != XDICT_HANDLE_CODE:
            raise DXFStructureError("ACAD_XDICTIONARY error.")
        return cls(tags[1].value)

    def load_resources(self, doc: 'Drawing') -> None:
        handle = self._xdict
        assert isinstance(handle, str)
        self._xdict = doc.entitydb.get(handle)

    def export_dxf(self, tagwriter: 'TagWriter') -> None:
        assert self._xdict is not None
        xdict = self._xdict
        handle = xdict if isinstance(xdict, str) else xdict.dxf.handle
        tagwriter.write_tag2(APP_DATA_MARKER, ACAD_XDICTIONARY)
        tagwriter.write_tag2(XDICT_HANDLE_CODE, handle)
        tagwriter.write_tag2(APP_DATA_MARKER, '}')

    def destroy(self) -> None:
        if self._xdict is not None:
            self._xdict.destroy()
            self._xdict = None

    def add_dictionary(self, name: str, doc: 'Drawing',
                       hard_owned: bool = False) -> 'DXFEntity':
        dictionary = self.dictionary
        new_dict = doc.objects.add_dictionary(
            owner=dictionary.dxf.hande,
            hard_owned=hard_owned,
        )
        dictionary[name] = new_dict
        return new_dict

    def add_placeholder(self, name: str,
                        doc: 'Drawing') -> 'DXFEntity':
        dictionary = self.dictionary
        placeholder = doc.objects.add_placeholder(dictionary.dxf.handle)
        dictionary[name] = placeholder
        return placeholder



# Purpose: read and write AutoCAD CTB files
# Created: 23.03.2010 for dxfwrite, added to ezdxf package on 2016-03-06
# Copyright (c) 2010-2019, Manfred Moitzi
# License: MIT License
# IMPORTANT: use only standard 7-Bit ascii code

from typing import Union, Tuple, Optional, BinaryIO, TextIO, Iterable, List, Any, Dict
from abc import abstractmethod
from io import StringIO
from array import array
from struct import pack
import zlib

END_STYLE_BUTT = 0
END_STYLE_SQUARE = 1
END_STYLE_ROUND = 2
END_STYLE_DIAMOND = 3
END_STYLE_OBJECT = 4

JOIN_STYLE_MITER = 0
JOIN_STYLE_BEVEL = 1
JOIN_STYLE_ROUND = 2
JOIN_STYLE_DIAMOND = 3
JOIN_STYLE_OBJECT = 5

FILL_STYLE_SOLID = 64
FILL_STYLE_CHECKERBOARD = 65
FILL_STYLE_CROSSHATCH = 66
FILL_STYLE_DIAMONDS = 67
FILL_STYLE_HORIZONTAL_BARS = 68
FILL_STYLE_SLANT_LEFT = 69
FILL_STYLE_SLANT_RIGHT = 70
FILL_STYLE_SQUARE_DOTS = 71
FILL_STYLE_VERICAL_BARS = 72
FILL_STYLE_OBJECT = 73

DITHERING_ON = 1  # bit coded color_policy
GRAYSCALE_ON = 2  # bit coded color_policy
NAMED_COLOR = 4  # bit coded color_policy

AUTOMATIC = 0
OBJECT_LINEWEIGHT = 0
OBJECT_LINETYPE = 31
OBJECT_COLOR = -1
OBJECT_COLOR2 = -1006632961

STYLE_COUNT = 255

DEFAULT_LINE_WEIGHTS = [
    0.00,  # 0
    0.05,  # 1
    0.09,  # 2
    0.10,  # 3
    0.13,  # 4
    0.15,  # 5
    0.18,  # 6
    0.20,  # 7
    0.25,  # 8
    0.30,  # 9
    0.35,  # 10
    0.40,  # 11
    0.45,  # 12
    0.50,  # 13
    0.53,  # 14
    0.60,  # 15
    0.65,  # 16
    0.70,  # 17
    0.80,  # 18
    0.90,  # 19
    1.00,  # 20
    1.06,  # 21
    1.20,  # 22
    1.40,  # 23
    1.58,  # 24
    2.00,  # 25
    2.11,  # 26
]

# color_type: (thx to Rammi)

# Take color from layer, ignore other bytes.
COLOR_BY_LAYER = 0xc0

# Take color from insertion, ignore other bytes
COLOR_BY_BLOCK = 0xc1

# RGB value, other bytes are R,G,B.
COLOR_RGB = 0xc2

# ACI, AutoCAD color index, other bytes are 0,0,index ???
COLOR_ACI = 0xc3


def color_name(index: int) -> str:
    return 'Color_%d' % (index + 1)


def get_bool(value: Union[str, bool]) -> bool:
    if isinstance(value, str):
        upperstr = value.upper()
        if upperstr == 'TRUE':
            value = True
        elif upperstr == 'FALSE':
            value = False
        else:
            raise ValueError("Unknown bool value '%s'." % str(value))
    return value


class PlotStyle:
    def __init__(self, index: int, data: dict = None, parent: 'PlotStyleTable' = None):
        data = data or {}
        self.parent = parent
        self.index = int(index)
        self.name = str(data.get('name', color_name(index)))
        self.localized_name = str(data.get('localized_name', color_name(index)))
        self.description = str(data.get('description', ""))
        # do not set _color, _mode_color or _color_policy directly
        # use set_color() method, and the properties dithering and grayscale
        self._color = int(data.get('color', OBJECT_COLOR))
        self._color_type = COLOR_RGB
        if self._color != OBJECT_COLOR:
            self._mode_color = int(data.get('mode_color', self._color))
        self._color_policy = int(data.get('color_policy', DITHERING_ON))
        self.physical_pen_number = int(data.get('physical_pen_number', AUTOMATIC))
        self.virtual_pen_number = int(data.get('virtual_pen_number', AUTOMATIC))
        self.screen = int(data.get('screen', 100))
        self.linepattern_size = float(data.get('linepattern_size', 0.5))
        self.linetype = int(data.get('linetype', OBJECT_LINETYPE))  # 0 .. 30
        self.adaptive_linetype = get_bool(data.get('adaptive_linetype', True))

        # lineweight index
        self.lineweight = int(data.get('lineweight', OBJECT_LINEWEIGHT))
        self.end_style = int(data.get('end_style', END_STYLE_OBJECT))
        self.join_style = int(data.get('join_style', JOIN_STYLE_OBJECT))
        self.fill_style = int(data.get('fill_style', FILL_STYLE_OBJECT))

    @property
    def color(self) -> Optional[Tuple[int, int, int]]:
        """  Get style color as ``(r, g, b)`` tuple or ``None``, if style has object color. """
        if self.has_object_color():
            return None  # object color
        else:
            return int2color(self._mode_color)[:3]

    @color.setter
    def color(self, rgb: Tuple[int, int, int]) -> None:
        """ Set color as RGB values. """
        r, g, b = rgb
        # when defining a user-color, `mode_color` represents the real true_color as (r, g, b) tuple and
        # color_type = COLOR_RGB (0xC2) as highest byte, the `color` value calculated for a user-color is not a
        # (r, g, b) tuple and has color_type = COLOR_ACI (0xC3) (sometimes), set for `color` the same value as for
        # `mode_color`, because AutoCAD corrects the `color` value by itself.
        self._mode_color = mode_color2int(r, g, b, color_type=self._color_type)
        self._color = self._mode_color

    @property
    def color_type(self):
        if self.has_object_color():
            return None  # object color
        else:
            return self._color_type

    @color_type.setter
    def color_type(self, value: int):
        self._color_type = value

    def set_object_color(self) -> None:
        """ Set color to object color. """
        self._color = OBJECT_COLOR
        self._mode_color = OBJECT_COLOR

    def set_lineweight(self, lineweight: float) -> None:
        """ Set `lineweight` in millimeters. Use ``0.0`` to set lineweight by object. """
        self.lineweight = self.parent.get_lineweight_index(lineweight)

    def get_lineweight(self) -> float:
        """ Returns the lineweight in millimeters or `0.0` for use entity lineweight. """
        return self.parent.lineweights[self.lineweight]

    def has_object_color(self) -> bool:
        """ ``True`` if style has object color. """
        return self._color in (OBJECT_COLOR, OBJECT_COLOR2)

    @property
    def aci(self) -> int:
        """ :ref:`ACI` in range from ``1`` to ``255``. Has no meaning for named plot styles. (int) """
        return self.index + 1

    @property
    def dithering(self) -> bool:
        """ Depending on the capabilities of your plotter, dithering approximates the colors with dot patterns.
        When this option is ``False``, the colors are mapped to the nearest color, resulting in a smaller range of
        colors when plotting.

        Dithering is available only whether you select the object’s color or assign a plot style color.

        """
        return bool(self._color_policy & DITHERING_ON)

    @dithering.setter
    def dithering(self, status: bool) -> None:
        if status:
            self._color_policy |= DITHERING_ON
        else:
            self._color_policy &= ~DITHERING_ON

    @property
    def grayscale(self) -> bool:
        """  Plot colors in grayscale. (bool) """
        return bool(self._color_policy & GRAYSCALE_ON)

    @grayscale.setter
    def grayscale(self, status: bool) -> None:
        if status:
            self._color_policy |= GRAYSCALE_ON
        else:
            self._color_policy &= ~GRAYSCALE_ON

    @property
    def named_color(self) -> bool:
        return bool(self._color_policy & NAMED_COLOR)

    @named_color.setter
    def named_color(self, status: bool) -> None:
        if status:
            self._color_policy |= NAMED_COLOR
        else:
            self._color_policy &= ~NAMED_COLOR

    def write(self, stream: TextIO) -> None:
        """ Write style data to file-like object `stream`. """
        index = self.index
        stream.write(' %d{\n' % index)
        stream.write('  name="%s\n' % self.name)
        stream.write('  localized_name="%s\n' % self.localized_name)
        stream.write('  description="%s\n' % self.description)
        stream.write('  color=%d\n' % self._color)
        if self._color != OBJECT_COLOR:
            stream.write('  mode_color=%d\n' % self._mode_color)
        stream.write('  color_policy=%d\n' % self._color_policy)
        stream.write('  physical_pen_number=%d\n' % self.physical_pen_number)
        stream.write('  virtual_pen_number=%d\n' % self.virtual_pen_number)
        stream.write('  screen=%d\n' % self.screen)
        stream.write('  linepattern_size=%s\n' % str(self.linepattern_size))
        stream.write('  linetype=%d\n' % self.linetype)
        stream.write('  adaptive_linetype=%s\n' % str(bool(self.adaptive_linetype)).upper())
        stream.write('  lineweight=%s\n' % str(self.lineweight))
        stream.write('  fill_style=%d\n' % self.fill_style)
        stream.write('  end_style=%d\n' % self.end_style)
        stream.write('  join_style=%d\n' % self.join_style)
        stream.write(' }\n')


class PlotStyleTable:
    """ PlotStyle container """

    def __init__(self, description: str = '', scale_factor: float = 1.0, apply_factor: bool = False):
        self.description = description
        self.scale_factor = scale_factor
        self.apply_factor = apply_factor

        # set custom_lineweight_display_units to 1 for showing lineweight in inch in AutoCAD CTB editor window, but
        # lineweight is always defined in mm
        self.custom_lineweight_display_units = 0
        self.lineweights = array('f', DEFAULT_LINE_WEIGHTS)

    def get_lineweight_index(self, lineweight: float) -> int:
        """ Get index of `lineweight` in the lineweight table or append `lineweight` to lineweight table. """
        try:
            return self.lineweights.index(lineweight)
        except ValueError:
            self.lineweights.append(lineweight)
            return len(self.lineweights) - 1

    def set_table_lineweight(self, index: int, lineweight: float) -> int:
        """
        Argument `index` is the lineweight table index, not the :ref:`ACI`.

        Args:
            index: lineweight table index = :attr:`PlotStyle.lineweight`
            lineweight: in millimeters

        """
        try:
            self.lineweights[index] = lineweight
            return index
        except IndexError:
            self.lineweights.append(lineweight)
            return len(self.lineweights) - 1

    def get_table_lineweight(self, index: int) -> float:
        """
        Returns lineweight in millimeters of lineweight table entry `index`.

        Args:
            index: lineweight table index = :attr:`PlotStyle.lineweight`

        Returns:
            lineweight in mm or ``0.0`` for use entity lineweight

        """
        return self.lineweights[index]

    def save(self, filename: str) -> None:
        """ Save CTB or STB file as `filename` to the file system. """
        with open(filename, 'wb') as stream:
            self.write(stream)

    def write(self, stream: BinaryIO) -> None:
        """ Compress and write the CTB or STB file to binary `stream`. """
        memfile = StringIO()
        self.write_content(memfile)
        memfile.write(chr(0))  # end of file
        body = memfile.getvalue()
        memfile.close()
        _compress(stream, body)

    @abstractmethod
    def write_content(self, stream: TextIO) -> None:
        pass

    def _write_lineweights(self, stream: TextIO) -> None:
        """ Write custom lineweight table to text `stream`. """
        stream.write('custom_lineweight_table{\n')
        for index, weight in enumerate(self.lineweights):
            stream.write(' %d=%.2f\n' % (index, weight))
        stream.write('}\n')

    def parse(self, text: str) -> None:
        """ Parse plot styles from CTB string `text`. """

        def set_lineweights(lineweights):
            if lineweights is None:
                return
            self.lineweights = array('f', [0.0] * len(lineweights))
            for key, value in lineweights.items():
                self.lineweights[int(key)] = float(value)

        parser = PlotStyleFileParser(text)
        self.description = parser.get('description', "")
        self.scale_factor = float(parser.get('scale_factor', 1.0))
        self.apply_factor = get_bool(parser.get('apply_factor', True))
        self.custom_lineweight_display_units = int(
            parser.get('custom_lineweight_display_units', 0))
        set_lineweights(parser.get('custom_lineweight_table', None))
        self.load_styles(parser.get('plot_style', {}))

    @abstractmethod
    def load_styles(self, styles):
        pass


class ColorDependentPlotStyles(PlotStyleTable):
    def __init__(self, description: str = '', scale_factor: float = 1.0, apply_factor: bool = False):
        super().__init__(description, scale_factor, apply_factor)
        self._styles = [PlotStyle(index, parent=self) for index in range(STYLE_COUNT)]  # type: List[PlotStyle]
        self._styles.insert(0, PlotStyle(256))  # 1-based array: insert dummy value for index 0

    def __getitem__(self, aci: int) -> PlotStyle:
        """ Returns :class:`PlotStyle` for :ref:`ACI` `aci`. """
        if 0 < aci < 256:
            return self._styles[aci]
        else:
            raise IndexError(aci)

    def __setitem__(self, aci: int, style: PlotStyle):
        """ Set plot `style` for `aci`. """
        if 0 < aci < 256:
            style.parent = self
            self._styles[aci] = style
        else:
            raise IndexError(aci)

    def __iter__(self) -> Iterable[PlotStyle]:
        """ Iterable of all plot styles. """
        return iter(self._styles[1:])

    def new_style(self, aci: int, data: dict = None) -> PlotStyle:
        """ Set `aci` to new attributes defined by `data` dict.

        Args:
            aci: :ref:`ACI`
            data: ``dict`` of :class:`PlotStyle` attributes: description, color, physical_pen_number,
                  virtual_pen_number, screen, linepattern_size, linetype, adaptive_linetype,
                  lineweight, end_style, join_style, fill_style

        """
        # ctb table index = aci - 1
        # ctb table starts with index 0, where aci == 0 means BYBLOCK
        style = PlotStyle(index=aci - 1, data=data)
        style.color_type = COLOR_RGB
        self[aci] = style
        return style

    def get_lineweight(self, aci: int):
        """ Returns the assigned lineweight for :class:`PlotStyle` `aci` in millimeter. """
        style = self[aci]
        lineweight = style.get_lineweight()
        if lineweight == 0.0:
            return None
        else:
            return lineweight

    def write_content(self, stream: TextIO) -> None:
        """ Write the CTB-file to text `stream`. """
        self._write_header(stream)
        self._write_aci_table(stream)
        self._write_plot_styles(stream)
        self._write_lineweights(stream)

    def _write_header(self, stream: TextIO) -> None:
        """ Write header values of CTB-file to text `stream`. """
        stream.write('description="%s\n' % self.description)
        stream.write('aci_table_available=TRUE\n')
        stream.write('scale_factor=%.1f\n' % self.scale_factor)
        stream.write('apply_factor=%s\n' % str(self.apply_factor).upper())
        stream.write('custom_lineweight_display_units=%s\n' % str(
            self.custom_lineweight_display_units))

    def _write_aci_table(self, stream: TextIO) -> None:
        """ Write AutoCAD Color Index table to text `stream`. """
        stream.write('aci_table{\n')
        for style in self:
            index = style.index
            stream.write(' %d="%s\n' % (index, color_name(index)))
        stream.write('}\n')

    def _write_plot_styles(self, stream: TextIO) -> None:
        """ Write user styles to text `stream`. """
        stream.write('plot_style{\n')
        for style in self:
            style.write(stream)
        stream.write('}\n')

    def load_styles(self, styles):
        for index, style in styles.items():
            index = int(index)
            style = PlotStyle(index, style)
            style.color_type = COLOR_RGB
            aci = index + 1
            self[aci] = style


class NamedPlotStyles(PlotStyleTable):
    def __init__(self, description: str = '', scale_factor: float = 1.0, apply_factor: bool = False):
        super().__init__(description, scale_factor, apply_factor)
        normal = PlotStyle(0, data={
            'name': 'Normal',
            'localized_name': 'Normal',
        })
        self._styles = {'Normal': normal}  # type: Dict[str, PlotStyle]

    def __iter__(self) -> Iterable[str]:
        """ Iterable of all plot style names. """
        return self.keys()

    def __getitem__(self, name: str) -> PlotStyle:
        """ Returns :class:`PlotStyle` by `name`. """
        return self._styles[name]

    def __delitem__(self, name: str) -> None:
        """ Delete plot style `name`. Plot style ``'Normal'`` is not deletable. """
        if name != 'Normal':
            del self._styles[name]
        else:
            raise ValueError("Can't delete plot style 'Normal'. ")

    def keys(self) -> Iterable[str]:
        """ Iterable of all plot style names. """
        keys = set(self._styles.keys())
        keys.discard('Normal')
        result = ['Normal']
        result.extend(sorted(keys))
        return iter(result)

    def items(self) -> Iterable[Tuple[str, PlotStyle]]:
        """ Iterable of all plot styles as (``name``, class:`PlotStyle`) tuples. """
        for key in self.keys():
            yield key, self._styles[key]

    def values(self) -> Iterable[PlotStyle]:
        """ Iterable of all class:`PlotStyle` objects. """
        for key, value in self.items():
            yield value

    def new_style(self, name: str, data: dict = None, localized_name: str = None) -> PlotStyle:
        """ Create new class:`PlotStyle` `name` by attribute dict `data`, replaces existing class:`PlotStyle` objects.

        Args:
            name: plot style name
            localized_name: name shown in plot style editor, uses `name` if ``None``
            data: ``dict`` of :class:`PlotStyle` attributes: description, color, physical_pen_number,
                  virtual_pen_number, screen, linepattern_size, linetype, adaptive_linetype,
                  lineweight, end_style, join_style, fill_style

        """
        if name.lower() == 'Normal':
            raise ValueError("Can't replace or modify plot style 'Normal'. ")
        data = data or {}
        data['name'] = name
        data['localized_name'] = localized_name or name
        index = len(self._styles)
        style = PlotStyle(index=index, data=data, parent=self)
        style.color_type = COLOR_ACI
        style.named_color = True
        self._styles[name] = style
        return style

    def get_lineweight(self, name: str):
        """ Returns the assigned lineweight for :class:`PlotStyle` `name` in millimeter. """
        style = self[name]
        lineweight = style.get_lineweight()
        if lineweight == 0.0:
            return None
        else:
            return lineweight

    def write_content(self, stream: TextIO) -> None:
        """ Write the STB-file to text `stream`. """
        self._write_header(stream)
        self._write_plot_styles(stream)
        self._write_lineweights(stream)

    def _write_header(self, stream: TextIO) -> None:
        """ Write header values of CTB-file to text `stream`. """
        stream.write('description="%s\n' % self.description)
        stream.write('aci_table_available=FALSE\n')
        stream.write('scale_factor=%.1f\n' % self.scale_factor)
        stream.write('apply_factor=%s\n' % str(self.apply_factor).upper())
        stream.write('custom_lineweight_display_units=%s\n' % str(
            self.custom_lineweight_display_units))

    def _write_plot_styles(self, stream: TextIO) -> None:
        """ Write user styles to text `stream`. """
        stream.write('plot_style{\n')
        for index, style in enumerate(self.values()):
            style.index = index
            style.write(stream)
        stream.write('}\n')

    def load_styles(self, styles):
        for index, style in styles.items():
            index = int(index)
            style = PlotStyle(index, style)
            style.color_type = COLOR_ACI
            self._styles[style.name] = style


def _read_ctb(stream: BinaryIO) -> ColorDependentPlotStyles:
    """ Read a CTB-file from from binary `stream`. """
    content = _decompress(stream)
    content = content.decode()
    styles = ColorDependentPlotStyles()
    styles.parse(content)
    return styles


def _read_stb(stream: BinaryIO) -> NamedPlotStyles:
    """ Read a STB-file from from binary `stream`. """
    content = _decompress(stream)
    content = content.decode()
    styles = NamedPlotStyles()
    styles.parse(content)
    return styles


def load(filename: str) -> Union[ColorDependentPlotStyles, NamedPlotStyles]:
    """ Load the CTB or STB file `filename` from file system. """

    with open(filename, 'rb') as stream:
        if filename.lower().endswith('.ctb'):
            table = _read_ctb(stream)
        elif filename.lower().endswith('.stb'):
            table = _read_stb(stream)
        else:
            raise ValueError('Invalid file type: "{}"'.format(filename))
    return table


def new_ctb() -> ColorDependentPlotStyles:
    """
    Create a new CTB file.

    .. versionchanged:: 0.10

        renamed from :func:`new`

    """
    return ColorDependentPlotStyles()


def new_stb() -> NamedPlotStyles:
    """ Create a new STB file. """
    return NamedPlotStyles()


def _decompress(stream: BinaryIO) -> bytes:
    """ Read and decompress the file content from binray `stream`. """
    content = stream.read()
    data = zlib.decompress(content[60:])  # type: bytes
    return data[:-1]  # truncate trailing \nul


def _compress(stream: BinaryIO, content: str):
    """ Compress `content` and write to binary `stream`. """

    def writestr(s):
        stream.write(s.encode())

    content = content.encode()
    comp_body = zlib.compress(content)
    adler_chksum = zlib.adler32(comp_body)
    writestr('PIAFILEVERSION_2.0,CTBVER1,compress\r\npmzlibcodec')
    stream.write(pack('LLL', adler_chksum, len(content), len(comp_body)))
    stream.write(comp_body)


class PlotStyleFileParser:
    """
    A very simple CTB/STB file parser. CTB/STB files are created by applications, so the file structure should be
    correct in the most cases.

    """

    def __init__(self, text: str):
        """
        :param str text: ctb content as string

        """
        self.data = {}
        for element, value in PlotStyleFileParser.iteritems(text):
            self.data[element] = value

    @staticmethod
    def iteritems(text: str):
        """
        iterate over all first level (start at col 0) elements

        """

        def get_name() -> str:
            """
            Get element name of line <line_index>.

            """
            line = lines[line_index]
            if line.endswith('{'):  # start of a list like 'plot_style{'
                name = line[:-1]
            else:  # simple name=value line
                name = line.split('=', 1)[0]
            return name.strip()

        def get_mapping() -> dict:
            """
            Get mapping of elements enclosed by { }.

            e. g. lineweigths, plot_styles, aci_table

            """
            nonlocal line_index

            def end_of_list():
                return lines[line_index].endswith('}')

            data = dict()
            while not end_of_list():
                name = get_name()
                value = get_value()  # get value or sub-list
                data[name] = value
            line_index += 1
            return data  # skip '}' - end of list

        def get_value() -> Union[str, dict]:
            """
            Get value of line <line_index> or the list that starts in line <line_index>.

            """
            nonlocal line_index
            line = lines[line_index]
            if line.endswith('{'):  # start of a list
                line_index += 1
                value = get_mapping()
            else:  # it's a simple name=value line
                value = line.split('=', 1)[1]  # type: str
                value = value.lstrip('"')  # strings look like this: name="value
                line_index += 1
            return value

        def skip_empty_lines():
            nonlocal line_index
            while line_index < len(lines) and len(lines[line_index]) == 0:
                line_index += 1

        lines = text.split('\n')
        line_index = 0
        while line_index < len(lines):
            name = get_name()
            value = get_value()
            yield (name, value)
            skip_empty_lines()

    def get(self, name: str, default: Any) -> Any:
        return self.data.get(name, default)


def int2color(color: int) -> Tuple[int, int, int, int]:
    """ Convert color integer value from CTB-file to ``(r, g, b, color_type) tuple. """
    # Take color from layer, ignore other bytes.
    color_type = (color & 0xff000000) >> 24
    red = (color & 0xff0000) >> 16
    green = (color & 0xff00) >> 8
    blue = color & 0xff
    return red, green, blue, color_type


def mode_color2int(red: int, green: int, blue: int, color_type=COLOR_RGB) -> int:
    """ Convert mode_color (r, g, b, color_type) tuple to integer. """
    return -color2int(red, green, blue, color_type)


def color2int(red: int, green: int, blue: int, color_type: int) -> int:
    """ Convert color (r, g, b, color_type) to integer. """
    return -((color_type << 24) + (red << 16) + (green << 8) + blue) & 0xffffffff



# Copyright (c) 2019-2020 Manfred Moitzi
# License: MIT License
from typing import TYPE_CHECKING, Iterable, List, Union
from contextlib import contextmanager
from ezdxf.lldxf.attributes import (
    DXFAttr, DXFAttributes, DefSubclass, XType, group_code_mapping,
)
from ezdxf.lldxf.const import (
    SUBCLASS_MARKER, DXF2000, DXFTypeError, DXF2013,
    DXFStructureError,
)
from ezdxf.lldxf.tags import Tags, DXFTag
from ezdxf.math import Matrix44
from ezdxf.tools import crypt
from .dxfentity import base_class, SubclassProcessor
from .dxfgfx import DXFGraphic, acdb_entity
from .factory import register_entity

if TYPE_CHECKING:
    from ezdxf.eztypes import TagWriter, DXFNamespace

__all__ = [
    'Body', 'Solid3d', 'Region', 'Surface', 'ExtrudedSurface', 'LoftedSurface',
    'RevolvedSurface', 'SweptSurface',
]

acdb_modeler_geometry = DefSubclass('AcDbModelerGeometry', {
    'version': DXFAttr(70, default=1),
    'flags': DXFAttr(290, dxfversion=DXF2013),
    'uid': DXFAttr(2, dxfversion=DXF2013),
})
acdb_modeler_geometry_group_codes = group_code_mapping(acdb_modeler_geometry)


# with R2013/AC1027 Modeler Geometry of ACIS data is stored in the ACDSDATA
# section as binary encoded information detection:
# group code 70, 1, 3 is missing
# group code 290, 2 present
#
#   0
# ACDSRECORD
#  90
# 1
#   2
# AcDbDs::ID
# 280
# 10
# 320
# 19B   <<< handle of associated 3DSOLID entity in model space
#   2
# ASM_Data
# 280
# 15
#  94
# 7197  <<< size in bytes ???
# 310
# 414349532042696E61727946696C6...


@register_entity
class Body(DXFGraphic):
    """ DXF BODY entity - container entity for embedded ACIS data. """
    DXFTYPE = 'BODY'
    DXFATTRIBS = DXFAttributes(base_class, acdb_entity, acdb_modeler_geometry)
    MIN_DXF_VERSION_FOR_EXPORT = DXF2000

    def __init__(self):
        super().__init__()
        self._acis_data: List[Union[str, bytes]] = []

    @property
    def acis_data(self) -> List[Union[str, bytes]]:
        """ Get ACIS text data as list of strings for DXF R2000 to DXF R2010 and binary encoded ACIS data for DXF R2013
        and later as list of bytes.
        """
        if self.has_binary_data:
            return self.doc.acdsdata.get_acis_data(self.dxf.handle)
        else:
            return self._acis_data

    @acis_data.setter
    def acis_data(self, lines: Iterable[str]):
        """ Set ACIS data as list of strings for DXF R2000 to DXF R2010. In case of DXF R2013 and later, setting ACIS
        data as binary data is not supported.
        """
        if self.has_binary_data:
            raise DXFTypeError(
                'Setting ACIS data not supported for DXF R2013 and later.')
        else:
            self._acis_data = list(lines)

    @property
    def has_binary_data(self):
        """ Returns ``True`` if ACIS data is of type ``List[bytes]``, ``False``
        if data is of type ``List[str]``.
        """
        if self.doc:
            return self.doc.dxfversion >= DXF2013
        else:
            return False

    def copy(self):
        """ Prevent copying. (internal interface)"""
        raise DXFTypeError('Copying of ACIS data not supported.')

    def load_dxf_attribs(
            self, processor: SubclassProcessor = None) -> 'DXFNamespace':
        """ Loading interface. (internal API)"""
        dxf = super().load_dxf_attribs(processor)
        if processor:
            # TODO: fast loader?
            processor.fast_load_dxfattribs(
                dxf, acdb_modeler_geometry_group_codes, 2, log=False)
            if not self.has_binary_data:
                self.load_acis_data(processor.subclasses[2])
        return dxf

    def load_acis_data(self, tags: Tags):
        """ Loading interface. (internal API)"""
        text_lines = tags2textlines(tag for tag in tags if tag.code in (1, 3))
        self.acis_data = crypt.decode(text_lines)

    def export_entity(self, tagwriter: 'TagWriter') -> None:
        """ Export entity specific data as DXF tags. (internal API)"""
        super().export_entity(tagwriter)
        tagwriter.write_tag2(SUBCLASS_MARKER, acdb_modeler_geometry.name)
        if tagwriter.dxfversion >= DXF2013:
            # ACIS data stored in the ACDSDATA section as binary encoded
            # information.
            if self.dxf.hasattr('version'):
                tagwriter.write_tag2(70, self.dxf.version)
            self.dxf.export_dxf_attribs(tagwriter, ['flags', 'uid'])
        else:
            # DXF R2000 - R2013 stores ACIS data as text in entity
            self.dxf.export_dxf_attribs(tagwriter, 'version')
            self.export_acis_data(tagwriter)

    def export_acis_data(self, tagwriter: 'TagWriter') -> None:
        """ Export ACIS data as DXF tags. (internal API)"""

        def cleanup(lines):
            for line in lines:
                yield line.rstrip().replace('\n', '')

        tags = Tags(textlines2tags(crypt.encode(cleanup(self.acis_data))))
        tagwriter.write_tags(tags)

    def set_text(self, text: str, sep: str = '\n') -> None:
        """ Set ACIS data from one string. """
        self.acis_data = text.split(sep)

    def tostring(self) -> str:
        """ Returns ACIS data as one string for DXF R2000 to R2010. """
        if self.has_binary_data:
            return ""
        else:
            return "\n".join(self.acis_data)

    def tobytes(self) -> bytes:
        """ Returns ACIS data as joined bytes for DXF R2013 and later. """
        if self.has_binary_data:
            return b"".join(self.acis_data)
        else:
            return b""

    def get_acis_data(self):
        """ Get the ACIS source code as a list of strings. """
        # for backward compatibility
        return self.acis_data

    def set_acis_data(self, text_lines: Iterable[str]) -> None:
        """ Set the ACIS source code as a list of strings **without** line
        endings.
        """
        # for backward compatibility
        self.acis_data = text_lines

    @contextmanager
    def edit_data(self) -> 'ModelerGeometry':
        # for backward compatibility
        data = ModelerGeometry(self)
        yield data
        self.acis_data = data.text_lines


class ModelerGeometry:
    # for backward compatibility
    def __init__(self, body: 'Body'):
        self.text_lines = body.acis_data

    def __str__(self) -> str:
        return "\n".join(self.text_lines)

    def set_text(self, text: str, sep: str = '\n') -> None:
        self.text_lines = text.split(sep)


def tags2textlines(tags: Iterable) -> Iterable[str]:
    """ Yields text lines from code 1 and 3 tags, code 1 starts a line following
    code 3 tags are appended to the line.
    """
    line = None
    for code, value in tags:
        if code == 1:
            if line is not None:
                yield line
            line = value
        elif code == 3:
            line += value
    if line is not None:
        yield line


def textlines2tags(lines: Iterable[str]) -> Iterable[DXFTag]:
    """ Yields text lines as DXFTags, splitting long lines (>255) int code 1
    and code 3 tags.
    """
    for line in lines:
        text = line[:255]
        tail = line[255:]
        yield DXFTag(1, text)
        while len(tail):
            text = tail[:255]
            tail = tail[255:]
            yield DXFTag(3, text)


@register_entity
class Region(Body):
    """ DXF REGION entity - container entity for embedded ACIS data. """
    DXFTYPE = 'REGION'


acdb_3dsolid = DefSubclass('AcDb3dSolid', {
    'history_handle': DXFAttr(350, default='0'),
})
acdb_3dsolid_group_codes = group_code_mapping(acdb_3dsolid)


@register_entity
class Solid3d(Body):
    """ DXF 3DSOLID entity - container entity for embedded ACIS data. """
    DXFTYPE = '3DSOLID'
    DXFATTRIBS = DXFAttributes(
        base_class, acdb_entity, acdb_modeler_geometry, acdb_3dsolid)

    def load_dxf_attribs(
            self, processor: SubclassProcessor = None) -> 'DXFNamespace':
        dxf = super().load_dxf_attribs(processor)
        if processor:
            processor.fast_load_dxfattribs(dxf, acdb_3dsolid_group_codes, 3)
        return dxf

    def export_entity(self, tagwriter: 'TagWriter') -> None:
        """ Export entity specific data as DXF tags. """
        # base class export is done by parent class
        super().export_entity(tagwriter)
        # AcDbEntity export is done by parent class
        # AcDbModelerGeometry export is done by parent class
        tagwriter.write_tag2(SUBCLASS_MARKER, acdb_3dsolid.name)
        self.dxf.export_dxf_attribs(tagwriter, 'history_handle')


def load_matrix(subclass: 'Tags', code: int) -> Matrix44:
    values = [tag.value for tag in subclass.find_all(code)]
    if len(values) != 16:
        raise DXFStructureError('Invalid transformation matrix.')
    return Matrix44(values)


def export_matrix(tagwriter: 'TagWriter', code: int, matrix: Matrix44) -> None:
    for value in list(matrix):
        tagwriter.write_tag2(code, value)


acdb_surface = DefSubclass('AcDbSurface', {
    'u_count': DXFAttr(71),
    'v_count': DXFAttr(72),
})
acdb_surface_group_codes = group_code_mapping(acdb_surface)


@register_entity
class Surface(Body):
    """ DXF SURFACE entity - container entity for embedded ACIS data. """
    DXFTYPE = 'SURFACE'
    DXFATTRIBS = DXFAttributes(
        base_class, acdb_entity, acdb_modeler_geometry, acdb_surface)

    def load_dxf_attribs(
            self, processor: SubclassProcessor = None) -> 'DXFNamespace':
        dxf = super().load_dxf_attribs(processor)
        if processor:
            # TODO: fast loader?
            processor.fast_load_dxfattribs(dxf, acdb_surface_group_codes, 3)
        return dxf

    def export_entity(self, tagwriter: 'TagWriter') -> None:
        """ Export entity specific data as DXF tags. """
        # base class export is done by parent class
        super().export_entity(tagwriter)
        # AcDbEntity export is done by parent class
        # AcDbModelerGeometry export is done by parent class
        tagwriter.write_tag2(SUBCLASS_MARKER, acdb_surface.name)
        self.dxf.export_dxf_attribs(tagwriter, ['u_count', 'v_count'])


acdb_extruded_surface = DefSubclass('AcDbExtrudedSurface', {
    'class_id': DXFAttr(90),
    'sweep_vector': DXFAttr(10, xtype=XType.point3d),
    # 16x group code 40: Transform matrix of extruded entity (16 floats;
    # row major format; default = identity matrix)
    'draft_angle': DXFAttr(42, default=0.),  # in radians
    'draft_start_distance': DXFAttr(43, default=0.),
    'draft_end_distance': DXFAttr(44, default=0.),
    'twist_angle': DXFAttr(45, default=0.),  # in radians?
    'scale_factor': DXFAttr(48, default=0.),
    'align_angle': DXFAttr(49, default=0.),  # in radians
    # 16x group code 46: Transform matrix of sweep entity (16 floats;
    # row major format; default = identity matrix)
    # 16x group code 47: Transform matrix of path entity (16 floats;
    # row major format; default = identity matrix)
    'solid': DXFAttr(290, default=0),  # bool

    # 0=No alignment; 1=Align sweep entity to path:
    'sweep_alignment_flags': DXFAttr(70, default=0),
    'unknown1': DXFAttr(71, default=0),

    # 2=Translate sweep entity to path; 3=Translate path to sweep entity:
    'align_start': DXFAttr(292, default=0),  # bool
    'bank': DXFAttr(293, default=0),  # bool
    'base_point_set': DXFAttr(294, default=0),  # bool
    'sweep_entity_transform_computed': DXFAttr(295, default=0),  # bool
    'path_entity_transform_computed': DXFAttr(296, default=0),  # bool
    'reference_vector_for_controlling_twist': DXFAttr(11, xtype=XType.point3d),
})
acdb_extruded_surface_group_codes = group_code_mapping(acdb_extruded_surface)


@register_entity
class ExtrudedSurface(Surface):
    """ DXF EXTRUDEDSURFACE entity - container entity for embedded ACIS data.
    """
    DXFTYPE = 'EXTRUDEDSURFACE'
    DXFATTRIBS = DXFAttributes(
        base_class, acdb_entity, acdb_modeler_geometry, acdb_surface,
        acdb_extruded_surface
    )

    def __init__(self):
        super().__init__()
        self.transformation_matrix_extruded_entity = Matrix44()
        self.sweep_entity_transformation_matrix = Matrix44()
        self.path_entity_transformation_matrix = Matrix44()

    def load_dxf_attribs(
            self, processor: SubclassProcessor = None) -> 'DXFNamespace':
        dxf = super().load_dxf_attribs(processor)
        if processor:
            processor.fast_load_dxfattribs(
                dxf, acdb_extruded_surface_group_codes, 4, log=False)
            self.load_matrices(processor.subclasses[4])
        return dxf

    def load_matrices(self, tags: Tags):
        self.transformation_matrix_extruded_entity = load_matrix(tags, code=40)
        self.sweep_entity_transformation_matrix = load_matrix(tags, code=46)
        self.path_entity_transformation_matrix = load_matrix(tags, code=47)

    def export_entity(self, tagwriter: 'TagWriter') -> None:
        """ Export entity specific data as DXF tags. """
        # base class export is done by parent class
        super().export_entity(tagwriter)
        # AcDbEntity export is done by parent class
        # AcDbModelerGeometry export is done by parent class
        tagwriter.write_tag2(SUBCLASS_MARKER, acdb_extruded_surface.name)
        self.dxf.export_dxf_attribs(tagwriter, ['class_id', 'sweep_vector'])
        export_matrix(tagwriter, code=40,
                      matrix=self.transformation_matrix_extruded_entity)
        self.dxf.export_dxf_attribs(tagwriter, [
            'draft_angle', 'draft_start_distance', 'draft_end_distance',
            'twist_angle', 'scale_factor', 'align_angle',
        ])
        export_matrix(tagwriter, code=46,
                      matrix=self.sweep_entity_transformation_matrix)
        export_matrix(tagwriter, code=47,
                      matrix=self.path_entity_transformation_matrix)
        self.dxf.export_dxf_attribs(tagwriter, [
            'solid', 'sweep_alignment_flags', 'unknown1', 'align_start', 'bank',
            'base_point_set', 'sweep_entity_transform_computed',
            'path_entity_transform_computed',
            'reference_vector_for_controlling_twist'
        ])


acdb_lofted_surface = DefSubclass('AcDbLoftedSurface', {
    # 16x group code 40: Transform matrix of loft entity (16 floats;
    # row major format; default = identity matrix)
    'plane_normal_lofting_type': DXFAttr(70),
    'start_draft_angle': DXFAttr(41, default=0.),  # in radians
    'end_draft_angle': DXFAttr(42, default=0.),  # in radians
    'start_draft_magnitude': DXFAttr(43, default=0.),
    'end_draft_magnitude': DXFAttr(44, default=0.),
    'arc_length_parameterization': DXFAttr(290, default=0),  # bool
    'no_twist': DXFAttr(291, default=1),  # true/false
    'align_direction': DXFAttr(292, default=1),  # bool
    'simple_surfaces': DXFAttr(293, default=1),  # bool
    'closed_surfaces': DXFAttr(294, default=0),  # bool
    'solid': DXFAttr(295, default=0),  # true/false
    'ruled_surface': DXFAttr(296, default=0),  # bool
    'virtual_guide': DXFAttr(297, default=0),  # bool
})
acdb_lofted_surface_group_codes = group_code_mapping(acdb_lofted_surface)


@register_entity
class LoftedSurface(Surface):
    """ DXF LOFTEDSURFACE entity - container entity for embedded ACIS data. """
    DXFTYPE = 'LOFTEDSURFACE'
    DXFATTRIBS = DXFAttributes(
        base_class, acdb_entity, acdb_modeler_geometry, acdb_surface,
        acdb_lofted_surface
    )

    def __init__(self):
        super().__init__()
        self.transformation_matrix_lofted_entity = Matrix44()

    def load_dxf_attribs(
            self, processor: SubclassProcessor = None) -> 'DXFNamespace':
        dxf = super().load_dxf_attribs(processor)
        if processor:
            processor.fast_load_dxfattribs(
                dxf, acdb_lofted_surface_group_codes, 4, log=False)
            self.load_matrices(processor.subclasses[4])
        return dxf

    def load_matrices(self, tags: Tags):
        self.transformation_matrix_lofted_entity = load_matrix(tags, code=40)

    def export_entity(self, tagwriter: 'TagWriter') -> None:
        """ Export entity specific data as DXF tags. """
        # base class export is done by parent class
        super().export_entity(tagwriter)
        # AcDbEntity export is done by parent class
        # AcDbModelerGeometry export is done by parent class
        tagwriter.write_tag2(SUBCLASS_MARKER, acdb_lofted_surface.name)
        export_matrix(tagwriter, code=40,
                      matrix=self.transformation_matrix_lofted_entity)
        self.dxf.export_dxf_attribs(
            tagwriter, acdb_lofted_surface.attribs.keys())


acdb_revolved_surface = DefSubclass('AcDbRevolvedSurface', {
    'class_id': DXFAttr(90, default=0.),
    'axis_point': DXFAttr(10, xtype=XType.point3d),
    'axis_vector': DXFAttr(11, xtype=XType.point3d),
    'revolve_angle': DXFAttr(40),  # in radians
    'start_angle': DXFAttr(41),  # in radians
    # 16x group code 42: Transform matrix of revolved entity (16 floats;
    # row major format; default = identity matrix)
    'draft_angle': DXFAttr(43),  # in radians
    'start_draft_distance': DXFAttr(44, default=0),
    'end_draft_distance': DXFAttr(45, default=0),
    'twist_angle': DXFAttr(46, default=0),  # in radians
    'solid': DXFAttr(290, default=0),  # bool
    'close_to_axis': DXFAttr(291, default=0),  # bool
})
acdb_revolved_surface_group_codes = group_code_mapping(acdb_revolved_surface)


@register_entity
class RevolvedSurface(Surface):
    """ DXF REVOLVEDSURFACE entity - container entity for embedded ACIS data.
    """
    DXFTYPE = 'REVOLVEDSURFACE'
    DXFATTRIBS = DXFAttributes(
        base_class, acdb_entity, acdb_modeler_geometry, acdb_surface,
        acdb_revolved_surface
    )

    def __init__(self):
        super().__init__()
        self.transformation_matrix_revolved_entity = Matrix44()

    def load_dxf_attribs(
            self, processor: SubclassProcessor = None) -> 'DXFNamespace':
        dxf = super().load_dxf_attribs(processor)
        if processor:
            processor.fast_load_dxfattribs(
                dxf, acdb_revolved_surface_group_codes, 4, log=False)
            self.load_matrices(processor.subclasses[4])
        return dxf

    def load_matrices(self, tags: Tags):
        self.transformation_matrix_revolved_entity = load_matrix(tags, code=42)

    def export_entity(self, tagwriter: 'TagWriter') -> None:
        """ Export entity specific data as DXF tags. """
        # base class export is done by parent class
        super().export_entity(tagwriter)
        # AcDbEntity export is done by parent class
        # AcDbModelerGeometry export is done by parent class
        tagwriter.write_tag2(SUBCLASS_MARKER, acdb_revolved_surface.name)
        self.dxf.export_dxf_attribs(tagwriter, [
            'class_id', 'axis_point', 'axis_vector', 'revolve_angle',
            'start_angle',
        ])
        export_matrix(tagwriter, code=42,
                      matrix=self.transformation_matrix_revolved_entity)
        self.dxf.export_dxf_attribs(tagwriter, [
            'draft_angle', 'start_draft_distance', 'end_draft_distance',
            'twist_angle', 'solid', 'close_to_axis',
        ])


acdb_swept_surface = DefSubclass('AcDbSweptSurface', {
    'swept_entity_id': DXFAttr(90),
    # 90: size of binary data (lost on saving)
    # 310: binary data  (lost on saving)
    'path_entity_id': DXFAttr(91),
    # 90: size of binary data  (lost on saving)
    # 310: binary data  (lost on saving)

    # 16x group code 40: Transform matrix of sweep entity (16 floats;
    # row major format; default = identity matrix)
    # 16x group code 41: Transform matrix of path entity (16 floats;
    # row major format; default = identity matrix)

    'draft_angle': DXFAttr(42),  # in radians
    'draft_start_distance': DXFAttr(43, default=0),
    'draft_end_distance': DXFAttr(44, default=0),
    'twist_angle': DXFAttr(45, default=0),  # in radians
    'scale_factor': DXFAttr(48, default=1),
    'align_angle': DXFAttr(49, default=0),  # in radians
    # don't know the meaning of this matrices
    # 16x group code 46: Transform matrix of sweep entity (16 floats;
    # row major format; default = identity matrix)
    # 16x group code 47: Transform matrix of path entity (16 floats;
    # row major format; default = identity matrix)
    'solid': DXFAttr(290, default=0),  # in radians

    # 0=No alignment; 1= align sweep entity to path:
    'sweep_alignment': DXFAttr(70, default=0),
    'unknown1': DXFAttr(71, default=0),

    # 2=Translate sweep entity to path; 3=Translate path to sweep entity:
    'align_start': DXFAttr(292, default=0),  # bool
    'bank': DXFAttr(293, default=0),  # bool
    'base_point_set': DXFAttr(294, default=0),  # bool
    'sweep_entity_transform_computed': DXFAttr(295, default=0),  # bool
    'path_entity_transform_computed': DXFAttr(296, default=0),  # bool
    'reference_vector_for_controlling_twist': DXFAttr(11, xtype=XType.point3d),
})
acdb_swept_surface_group_codes = group_code_mapping(acdb_swept_surface)


@register_entity
class SweptSurface(Surface):
    """ DXF SWEPTSURFACE entity - container entity for embedded ACIS data. """
    DXFTYPE = 'SWEPTSURFACE'
    DXFATTRIBS = DXFAttributes(
        base_class, acdb_entity, acdb_modeler_geometry, acdb_surface,
        acdb_swept_surface
    )

    def __init__(self):
        super().__init__()
        self.transformation_matrix_sweep_entity = Matrix44()
        self.transformation_matrix_path_entity = Matrix44()
        self.sweep_entity_transformation_matrix = Matrix44()
        self.path_entity_transformation_matrix = Matrix44()

    def load_dxf_attribs(
            self, processor: SubclassProcessor = None) -> 'DXFNamespace':
        dxf = super().load_dxf_attribs(processor)
        if processor:
            processor.fast_load_dxfattribs(
                dxf, acdb_swept_surface_group_codes, 4, log=False)
            self.load_matrices(processor.subclasses[4])
        return dxf

    def load_matrices(self, tags: Tags):
        self.transformation_matrix_sweep_entity = load_matrix(tags, code=40)
        self.transformation_matrix_path_entity = load_matrix(tags, code=41)
        self.sweep_entity_transformation_matrix = load_matrix(tags, code=46)
        self.path_entity_transformation_matrix = load_matrix(tags, code=47)

    def export_entity(self, tagwriter: 'TagWriter') -> None:
        """ Export entity specific data as DXF tags. """
        # base class export is done by parent class
        super().export_entity(tagwriter)
        # AcDbEntity export is done by parent class
        # AcDbModelerGeometry export is done by parent class
        tagwriter.write_tag2(SUBCLASS_MARKER, acdb_swept_surface.name)
        self.dxf.export_dxf_attribs(tagwriter, [
            'swept_entity_id', 'path_entity_id',
        ])
        export_matrix(tagwriter, code=40,
                      matrix=self.transformation_matrix_sweep_entity)
        export_matrix(tagwriter, code=41,
                      matrix=self.transformation_matrix_path_entity)
        self.dxf.export_dxf_attribs(tagwriter, [
            'draft_angle', 'draft_start_distance', 'draft_end_distance',
            'twist_angle', 'scale_factor', 'align_angle'
        ])

        export_matrix(tagwriter, code=46,
                      matrix=self.sweep_entity_transformation_matrix)
        export_matrix(tagwriter, code=47,
                      matrix=self.path_entity_transformation_matrix)
        self.dxf.export_dxf_attribs(tagwriter, [
            'solid', 'sweep_alignment', 'unknown1', 'align_start', 'bank',
            'base_point_set', 'sweep_entity_transform_computed',
            'path_entity_transform_computed',
            'reference_vector_for_controlling_twist'
        ])



# Copyright (c) 2019-2020, Manfred Moitzi
# License: MIT-License
from typing import TYPE_CHECKING
import copy
from ezdxf.lldxf import validator
from ezdxf.lldxf.const import SUBCLASS_MARKER, DXF2000
from ezdxf.lldxf.attributes import (
    DXFAttributes, DefSubclass, DXFAttr, group_code_mapping,
)
from ezdxf.lldxf.tags import Tags
from .dxfentity import base_class, SubclassProcessor
from .dxfobj import DXFObject
from .factory import register_entity

if TYPE_CHECKING:
    from ezdxf.eztypes import TagWriter, DXFNamespace

__all__ = ['VisualStyle']

acdb_visualstyle = DefSubclass('AcDbVisualStyle', {
    'description': DXFAttr(2),
    # Style type:
    # 0 = Flat
    # 1 = FlatWithEdges
    # 2 = Gouraud
    # 3 = GouraudWithEdges
    # 4 = 2dWireframe
    # 5 = Wireframe
    # 6 = Hidden
    # 7 = Basic
    # 8 = Realistic
    # 9 = Conceptual
    # 10 = Modeling
    # 11 = Dim
    # 12 = Brighten
    # 13 = Thicken
    # 14 = Linepattern
    # 15 = Facepattern
    # 16 = ColorChange
    # 20 = JitterOff
    # 21 = OverhangOff
    # 22 = EdgeColorOff
    # 23 = Shades of Gray
    # 24 = Sketchy
    # 25 = X-Ray
    # 26 = Shaded with edges
    # 27 = Shaded
    'style_type': DXFAttr(70, validator=validator.is_in_integer_range(0, 28)),

    # Face lighting type:
    # 0 = Invisible
    # 1 = Visible
    # 2 = Phong
    # 3 = Gooch
    'face_lighting_model': DXFAttr(
        71, validator=validator.is_in_integer_range(0, 4)),

    # Face lighting quality:
    # 0 = No lighting
    # 1 = Per face lighting
    # 2 = Per vertex lighting
    # 3 = Unknown
    'face_lighting_quality': DXFAttr(
        72, validator=validator.is_in_integer_range(0, 4)),

    # Face color mode:
    # 0 = No color
    # 1 = Object color
    # 2 = Background color
    # 3 = Custom color
    # 4 = Mono color
    # 5 = Tinted
    # 6 = Desaturated
    'face_color_mode': DXFAttr(
        73, validator=validator.is_in_integer_range(0, 6)),

    # Face modifiers:
    # 0 = No modifiers
    # 1 = Opacity
    # 2 = Specular
    # 3 = Unknown
    'face_modifiers': DXFAttr(
        90, validator=validator.is_in_integer_range(0, 4)),

    'face_opacity_level': DXFAttr(40),
    'face_specular_level': DXFAttr(41),
    'color1': DXFAttr(62),
    'color2': DXFAttr(63),
    'face_style_mono_color': DXFAttr(421),

    # Edge style model:
    # 0 = No edges
    # 1 = Isolines
    # 2 = Facet edges
    'edge_style_model': DXFAttr(
        74, validator=validator.is_in_integer_range(0, 3)),
    'edge_style': DXFAttr(91),
    'edge_intersection_color': DXFAttr(64),
    'edge_obscured_color': DXFAttr(65),
    'edge_obscured_linetype': DXFAttr(75),
    'edge_intersection_linetype': DXFAttr(175),
    'edge_crease_angle': DXFAttr(42),
    'edge_modifiers': DXFAttr(92),
    'edge_color': DXFAttr(66),
    'edge_opacity_level': DXFAttr(43),
    'edge_width': DXFAttr(76),
    'edge_overhang': DXFAttr(77),
    'edge_jitter': DXFAttr(78),
    'edge_silhouette_color': DXFAttr(67),
    'edge_silhouette_width': DXFAttr(79),
    'edge_halo_gap': DXFAttr(170),
    'edge_isoline_count': DXFAttr(171),
    'edge_hide_precision': DXFAttr(290),  # flag
    'edge_style_apply': DXFAttr(174),  # flag
    'style_display_settings': DXFAttr(93),
    'brightness': DXFAttr(44),
    'shadow_type': DXFAttr(173),
    'unknown1': DXFAttr(177),  # required if xdata is present?
    'internal_use_only_flag': DXFAttr(291),  # visual style only use internal
    # Xdata must follow tag 291 (AutoCAD -> 'Xdata wasn't read' error)
    # 70: Xdata count (count of tag groups)
    # any code, value: multiple tags build a tag group
    # 176, 1: end of group marker
    # e.g. (291, 0) (70, 2) (62, 7) (420, 16777215) (176, 1) (90, 1) (176, 1)
})
acdb_visualstyle_group_codes = group_code_mapping(acdb_visualstyle)


# undocumented Xdata in DXF R2018
@register_entity
class VisualStyle(DXFObject):
    """ DXF VISUALSTYLE entity """
    DXFTYPE = 'VISUALSTYLE'
    DXFATTRIBS = DXFAttributes(base_class, acdb_visualstyle)
    MIN_DXF_VERSION_FOR_EXPORT = DXF2000  # official supported in R2007

    def __init__(self):
        super().__init__()
        self.acad_xdata = None  # to preserve AutoCAD xdata

    def _copy_data(self, entity: 'VisualStyle') -> None:
        """ Copy acad internal data. """
        entity.acad_xdata = copy.deepcopy(self.acad_xdata)

    def load_dxf_attribs(
            self, processor: SubclassProcessor = None) -> 'DXFNamespace':
        dxf = super().load_dxf_attribs(processor)
        if processor:
            tags = processor.subclass_by_index(1)
            self.acad_xdata = self.store_acad_xdata(tags)
            processor.fast_load_dxfattribs(
                dxf, acdb_visualstyle_group_codes, subclass=tags)
        return dxf

    @staticmethod
    def store_acad_xdata(tags: Tags):
        try:
            index = tags.tag_index(291)
        except IndexError:
            return None
        else:  # store tags after 291
            index += 1
            xdata = tags[index:]
            del tags[index:]  # remove xdata from subclass
            return xdata

    def export_entity(self, tagwriter: 'TagWriter') -> None:
        """ Export entity specific data as DXF tags. """
        super().export_entity(tagwriter)
        tagwriter.write_tag2(SUBCLASS_MARKER, acdb_visualstyle.name)
        self.dxf.export_dxf_attribs(tagwriter, [
            'description', 'style_type', 'face_lighting_model',
            'face_lighting_quality', 'face_color_mode', 'face_modifiers',
            'face_opacity_level', 'face_specular_level', 'color1', 'color2',
            'face_style_mono_color', 'edge_style_model', 'edge_style',
            'edge_intersection_color', 'edge_obscured_color',
            'edge_obscured_linetype', 'edge_intersection_linetype',
            'edge_crease_angle', 'edge_modifiers', 'edge_color',
            'edge_opacity_level', 'edge_width', 'edge_overhang', 'edge_jitter',
            'edge_silhouette_color', 'edge_silhouette_width', 'edge_halo_gap',
            'edge_isoline_count', 'edge_hide_precision', 'edge_style_apply',
            'style_display_settings', 'brightness', 'shadow_type', 'unknown1',
            'internal_use_only_flag',
        ])
        if self.acad_xdata:
            tagwriter.write_tags(self.acad_xdata)
